/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath;

import org.intellij.lang.xpath.completion.CompletionLists;
import org.intellij.lang.xpath.psi.XPathFunction;
import org.intellij.lang.xpath.validation.XPathAnnotator;

import com.intellij.lang.BracePair;
import com.intellij.lang.Language;
import com.intellij.lang.PairedBraceMatcher;
import com.intellij.lang.ParserDefinition;
import com.intellij.lang.annotation.Annotator;
import com.intellij.lang.cacheBuilder.WordsScanner;
import com.intellij.lang.cacheBuilder.SimpleWordsScanner;
import com.intellij.lang.findUsages.FindUsagesProvider;
import com.intellij.lang.refactoring.NamesValidator;
import com.intellij.navigation.ItemPresentation;
import com.intellij.navigation.NavigationItem;
import com.intellij.openapi.fileTypes.SyntaxHighlighter;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiNamedElement;
import com.intellij.psi.tree.TokenSet;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public final class XPathLanguage extends Language {
    public static final String ID = "XPath";

    private static final Annotator ANNOTATOR = new XPathAnnotator();
    private static final XPathParserDefinition PARSER_DEFINITION = new XPathParserDefinition();
    private static final MyFindUsagesProvider FIND_USAGES_PROVIDER = new MyFindUsagesProvider();
    private static final MyNamesValidator NAMES_VALIDATOR = new MyNamesValidator();

    XPathLanguage() {
        super(ID);
    }

    @NotNull
    public ParserDefinition getParserDefinition() {
        return PARSER_DEFINITION;
    }

    @NotNull
    public SyntaxHighlighter getSyntaxHighlighter(Project project, VirtualFile file) {
        return new XPathHighlighter();
    }

    @Nullable
    public PairedBraceMatcher getPairedBraceMatcher() {
        return new MyPairedBraceMatcher();
    }

    @NotNull
    public TokenSet getReadableTextContainerElements() {
        return TokenSet.orSet(TokenSet.create(XPathTokenTypes.STRING_LITERAL));
    }

    public Annotator getAnnotator() {
        return ANNOTATOR;
    }

    @NotNull
    public FindUsagesProvider getFindUsagesProvider() {
        return FIND_USAGES_PROVIDER;
    }

    @NotNull
    public NamesValidator getNamesValidator() {
        return NAMES_VALIDATOR;
    }

    private static class MyPairedBraceMatcher implements PairedBraceMatcher {
        public BracePair[] getPairs() {
            return new BracePair[] {
              new BracePair('(', XPathTokenTypes.LPAREN, ')', XPathTokenTypes.RPAREN, false),
              new BracePair('[', XPathTokenTypes.LBRACKET, ']', XPathTokenTypes.RBRACKET, false),
            };
        }
    }

    private static class MyFindUsagesProvider implements FindUsagesProvider {
        @Nullable
        public WordsScanner getWordsScanner() {
            return new SimpleWordsScanner();
        }

        public boolean canFindUsagesFor(PsiElement psiElement) {
            return psiElement instanceof XPathFunction;
        }

        @Nullable
        public String getHelpId(PsiElement psiElement) {
            return null;
        }

        @NotNull
        public String getType(PsiElement element) {
            return element instanceof XPathFunction ? "function" : "unknown";
        }

        @NotNull
        public String getDescriptiveName(PsiElement element) {
            if (element instanceof PsiNamedElement) {
                final String name = ((PsiNamedElement)element).getName();
                if (name != null) return name;
            }
            return element.toString();
        }

        @NotNull
        public String getNodeText(PsiElement element, boolean useFullName) {
            if (useFullName) {
                if (element instanceof NavigationItem) {
                    final NavigationItem navigationItem = ((NavigationItem)element);
                    final ItemPresentation presentation = navigationItem.getPresentation();
                    if (presentation != null && presentation.getPresentableText() != null) {
                        return presentation.getPresentableText();
                    }
                    final String name = navigationItem.getName();
                    if (name != null) {
                        return name;
                    }
                }
            }
            if (element instanceof PsiNamedElement) {
                final String name = ((PsiNamedElement)element).getName();
                if (name != null) return name;
            }
            return element.toString();
        }
    }

    private static class MyNamesValidator implements NamesValidator {
        private final XPathLexer xPathLexer = new XPathLexer();

        public synchronized boolean isIdentifier(String text, Project project) {
            xPathLexer.start(text.toCharArray());
            assert xPathLexer.getState() == 0;

            boolean b = xPathLexer.getTokenType() == XPathTokenTypes.NCNAME;
            xPathLexer.advance();

            if (xPathLexer.getTokenType() == null) {
                return b;
            } else if (xPathLexer.getTokenType() == XPathTokenTypes.COL) {
                xPathLexer.advance();
                b = xPathLexer.getTokenType() == XPathTokenTypes.NCNAME;
                xPathLexer.advance();
                return b && xPathLexer.getTokenType() == null;
            }

            return false;
        }

        public boolean isKeyword(String text, Project project) {
            return CompletionLists.AXIS_NAMES.contains(text) || CompletionLists.NODE_TYPE_FUNCS.contains(text) || CompletionLists.OPERATORS.contains(text);
        }
    }
}
