/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath;

import org.intellij.lang.xpath.completion.XPathCompletionData;
import org.intellij.lang.xpath.psi.XPathElement;
import org.intellij.lang.xpath.psi.XPathFunction;
import org.intellij.lang.xpath.psi.XPathFunctionCall;
import org.intellij.lang.xpath.validation.inspections.*;

import com.intellij.codeInsight.completion.CompletionUtil;
import com.intellij.codeInsight.editorActions.TypedHandler;
import com.intellij.codeInsight.hint.ShowParameterInfoHandler;
import com.intellij.codeInsight.hint.api.*;
import com.intellij.codeInspection.InspectionToolProvider;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.fileTypes.FileTypeManager;
import com.intellij.openapi.options.colors.ColorSettingsPages;
import com.intellij.psi.filters.TrueFilter;
import com.intellij.psi.tree.IElementType;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

public class XPathSupportLoader implements ApplicationComponent, InspectionToolProvider {
    private static final boolean DBG_MODE = Boolean.getBoolean("xpath-lang.register-file-type");

    @NotNull
    @NonNls
    public String getComponentName() {
        return "XPath Support Loader";
    }
    
    public void initComponent() {
        if (true || DBG_MODE) {
            ApplicationManager.getApplication().runWriteAction(new Runnable() {
                public void run() {
                    final String[] extensions = new String[]{ XPathFileType.XPATH.getDefaultExtension() };
                    FileTypeManager.getInstance().registerFileType(XPathFileType.XPATH, extensions);
                }
            });
        }
        XPathFileType.XPATH.getLanguage().associateFileType(XPathFileType.XPATH);

        CompletionUtil.registerCompletionData(XPathFileType.XPATH, new XPathCompletionData(TrueFilter.INSTANCE, XPathElement.class));
        TypedHandler.registerQuoteHandler(XPathFileType.XPATH, new TypedHandler.SimpleTokenSetQuoteHandler(new IElementType[]{ XPathTokenTypes.STRING_LITERAL }));
        ColorSettingsPages.getInstance().registerPage(new XPathColorSettingsPage());
        ShowParameterInfoHandler.register(XPathFileType.XPATH.getLanguage(), new MyParameterInfoProvider());
    }

    public void disposeComponent() {
    }

    public Class<? extends XPathInspection>[] getInspectionClasses() {
        //noinspection unchecked
        return new Class[]{
                CheckNodeTest.class,
                ImplicitTypeConversion.class,
                RedundantTypeConversion.class,
                IndexZeroPredicate.class,
        };
    }

    private static class MyParameterInfoProvider implements ParameterInfoProvider {
        @NotNull
        public ParameterInfoHandler<XPathFunctionCall, XPathFunction>[] getHandlers() {
            //noinspection unchecked
            return new ParameterInfoHandler[] {
                    new XPathParameterInfoHandler()
            };
        }
    }
}
