/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.completion;

import org.intellij.lang.xpath.psi.PrefixedName;
import org.intellij.lang.xpath.psi.XPathFunctionCall;
import org.intellij.lang.xpath.psi.XPathNodeTest;
import org.intellij.lang.xpath.psi.XPathVariableReference;
import org.intellij.lang.xpath.psi.impl.XPathElementImpl;

import com.intellij.codeInsight.CodeInsightSettings;
import com.intellij.codeInsight.completion.CompletionContext;
import com.intellij.codeInsight.completion.CompletionData;
import com.intellij.codeInsight.completion.CompletionVariant;
import com.intellij.codeInsight.lookup.LookupItem;
import com.intellij.ide.IconUtilEx;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.Iconable;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiElementFactory;
import com.intellij.psi.PsiReference;
import com.intellij.psi.filters.ContextGetter;
import com.intellij.psi.filters.ElementFilter;
import com.intellij.psi.filters.TrueFilter;
import com.intellij.util.ArrayUtil;

import javax.swing.*;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

public class XPathCompletionData extends CompletionData {
    private static final Logger LOG = Logger.getInstance(XPathCompletionData.class.getName());

    public static final XPathInsertHandler INSERT_HANDLER = new XPathInsertHandler();

    public XPathCompletionData(ElementFilter instance, Class<?> scopeClass) {
        declareFinalScope(XPathElementImpl.class);

        final CompletionVariant completionvariant = new MyCompletionVariant(instance);

        completionvariant.includeScopeClass(scopeClass, true);
        completionvariant.addCompletionFilterOnElement(TrueFilter.INSTANCE);
        completionvariant.setInsertHandler(INSERT_HANDLER);

        completionvariant.addCompletion(new ContextGetter() {
            public Object[] get(PsiElement _element, CompletionContext completionContext) {
                PsiElement element = getContext(_element, completionContext.offset);

                LOG.debug("element = " + element);
                final PsiElement parent = element.getParent();

                final Collection<Lookup> lookups;
                if (parent instanceof XPathNodeTest) {
                    lookups = CompletionLists.getNodeTestCompletions((XPathNodeTest)parent);
                } else if (parent instanceof XPathFunctionCall) {
                    lookups = CompletionLists.getFunctionCompletions((XPathFunctionCall)parent);
                    lookups.addAll(CompletionLists.getNodeTypeCompletions());
                } else if (parent instanceof XPathVariableReference) {
                    lookups = CompletionLists.getVariableCompletions((XPathVariableReference)parent);
                } else {
                    return ArrayUtil.EMPTY_OBJECT_ARRAY;
                }
                return lookups.toArray(new Object[lookups.size()]);
            }
        });
        registerVariant(completionvariant);
    }

    private static class MyCompletionVariant extends CompletionVariant {
        public MyCompletionVariant(ElementFilter instance) {
            super(instance);
        }

        public void addKeywords(PsiElementFactory psiElementFactory, Set<LookupItem> set, CompletionContext completionContext, PsiElement element) {
            LOG.debug("set (before keywords) = " + set);

            final Set<LookupItem> copy = new HashSet<LookupItem>(set);
            super.addKeywords(psiElementFactory, set, completionContext, element);

            postProcess(set, copy);

            LOG.debug("set (after keywords) = " + set);
        }

        private void postProcess(Set<LookupItem> set, Set<LookupItem> copy) {
            final CodeInsightSettings codeinsightsettings = CodeInsightSettings.getInstance();
            final Icon emptyIcon = IconUtilEx.getEmptyIcon(codeinsightsettings.SHOW_SIGNATURES_IN_LOOKUPS);

            for (LookupItem item : set) {
                final Object obj = item.getObject();
                if (obj instanceof Iconable) {
                    final Icon icon = ((Iconable)obj).getIcon(Iconable.ICON_FLAG_CLOSED);
                    if (icon != null) {
                        if (icon.getIconWidth() < emptyIcon.getIconWidth()) {
                            item.setAttribute(LookupItem.ICON_ATTR, IconUtilEx.createLayeredIcon(emptyIcon, icon));
                        } else {
                            item.setAttribute(LookupItem.ICON_ATTR, icon);
                        }
                    }
                }
            }

            // I've seen this working to get rid of the useless and redundant word-completion lookups, but it doesn't work consistently
            removeDuplicates(copy, set, true);
        }
    }

    public void completeReference(PsiReference psiReference, Set<LookupItem> set, CompletionContext completionContext, PsiElement psiElement) {
        super.completeReference(psiReference, set, completionContext, psiElement);
        XPathCompletionData.filterLocalCompletions(set, true);
    }

    private static void filterLocalCompletions(Set<LookupItem> set, boolean keepLocal) {
        for (Iterator<LookupItem> it = set.iterator(); it.hasNext();) {
            final LookupItem item = it.next();
            if (keepLocal) {
                if (item.getInsertHandler() != INSERT_HANDLER) {
                    it.remove();
                }
            } else if (item.getInsertHandler() == INSERT_HANDLER) {
                it.remove();
            }
        }
    }

    @SuppressWarnings({ "RawUseOfParameterizedType", "MethodOnlyUsedFromNestedClass"})
    private static void removeDuplicates(Set<LookupItem> copy, Set<LookupItem> set, boolean preferLocal) {
        for (LookupItem i : copy) {
            final String name = i.getLookupString();
            for (Iterator<LookupItem> it = set.iterator(); it.hasNext();) {
                final LookupItem item = it.next();
                if (item != i && item.getLookupString().equals(name)) {
                    if (preferLocal) {
                        if (item.getInsertHandler() != INSERT_HANDLER) {
                            it.remove();
                        }
                    } else if (item.getInsertHandler() == INSERT_HANDLER) {
                        it.remove();
                    }
                }
            }
        }
    }

    public String findPrefix(PsiElement element, int i) {

        String prefix = super.findPrefix(element, i);

        final PsiElement context = getContext(element, i);

        if (context.getParent() instanceof XPathVariableReference) {
            final String text = context.getText();
            prefix = "$" + text.substring(0, text.indexOf(CompletionLists.INTELLIJ_IDEA_RULEZ));
        }

        if (context.getParent() instanceof XPathNodeTest) {
            final XPathNodeTest nodeTest = ((XPathNodeTest)context.getParent());
            if (nodeTest.isNameTest()) {
                final PrefixedName prefixedName = nodeTest.getQName();
                assert prefixedName != null;
                final String p = prefixedName.getPrefix();

                int endIndex = prefixedName.getLocalName().indexOf(CompletionLists.INTELLIJ_IDEA_RULEZ);
                if (endIndex != -1) {
                    prefix = prefixedName.getLocalName().substring(0, endIndex);
                } else if (p != null) {
                    endIndex = p.indexOf(CompletionLists.INTELLIJ_IDEA_RULEZ);
                    if (endIndex != -1) {
                        prefix = p.substring(0, endIndex);
                    }
                }
            }
        }

        LOG.debug("prefix = " + prefix);
        return prefix;
    }

    protected PsiElement getContext(PsiElement element, int i) {
        return element;
    }
}
