/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.context;

import org.intellij.lang.xpath.XPathFile;
import org.intellij.lang.xpath.context.functions.DefaultFunctionContext;
import org.intellij.lang.xpath.context.functions.Function;
import org.intellij.lang.xpath.context.functions.FunctionContext;
import org.intellij.lang.xpath.psi.*;
import org.intellij.lang.xpath.validation.inspections.quickfix.XPathQuickFixFactory;

import com.intellij.openapi.util.Key;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlElement;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.xml.namespace.QName;
import java.util.Set;

public abstract class ContextProvider {
    /**
     * ContextProvider instance, must be available (aka @NotNull) at every XPathFile instance by getCopyableUserData().
     * It gets lost with putUserData()!
     */
    private static final Key<ContextProvider> KEY = Key.create("CONTEXT_PROVIDER");

    private DefaultFunctionContext myFunctionContext;

    protected ContextProvider() {
    }

    protected ContextProvider(XPathFile element) {
        attachTo(element);
    }

    @NotNull
    public abstract ContextType getContextType();

    @Nullable
    public abstract XmlElement getContextElement();

    @Nullable
    public abstract NamespaceContext getNamespaceContext();

    @Nullable
    public abstract VariableContext getVariableContext();

    @NotNull
    public synchronized FunctionContext getFunctionContext() {
        if (myFunctionContext == null) {
            myFunctionContext = DefaultFunctionContext.getInstance(getContextType());
        }
        return myFunctionContext;
    }

    @NotNull
    public synchronized XPathQuickFixFactory getQuickFixFactory() {
        return XPathQuickFixFactoryImpl.INSTANCE;
    }

    @Nullable
    public abstract Set<QName> getAttributes(boolean forValidation);

    @Nullable
    public abstract Set<QName> getElements(boolean forValidation);

    public final void attachTo(PsiFile file) {
        assert file instanceof XPathFile;
        file.putCopyableUserData(KEY, this);
    }

    public final void attachTo(XmlElement context) {
        context.putCopyableUserData(KEY, this);
    }

    @SuppressWarnings({ "ClassReferencesSubclass" })
    @NotNull
    public static ContextProvider getContextProvider(PsiFile psiFile) {
        ContextProvider provider = psiFile.getCopyableUserData(KEY);
        if (provider != null) {
            return provider;
        } else {
            final PsiElement context = psiFile.getContext();
            if (context != null) {
                provider = context.getCopyableUserData(KEY);
                if (provider != null) {
                    return provider;
                }
            }
        }
        return new DefaultProvider(PsiTreeUtil.getContextOfType(psiFile, XmlElement.class, true));
    }

    @SuppressWarnings({ "ClassReferencesSubclass" })
    @NotNull
    public static ContextProvider getContextProvider(PsiElement element) {
        return element instanceof XPathElement ?
                getContextProvider(element instanceof XPathFile ?
                        (PsiFile)element :
                        element.getContainingFile()) :
                new DefaultProvider(PsiTreeUtil.getParentOfType(element, XmlElement.class, false));
    }

    @NotNull
    public XPathType getFunctionType(XPathFunctionCall call) {
        final Function function = getFunctionContext().getFunctions().get(getQName(call.getQName(), call));
        return function != null ? function.returnType : XPathType.UNKNOWN;
    }

    public PsiFile[] getRelatedFiles(XPathFile file) {
        return PsiFile.EMPTY_ARRAY;
    }

    @NotNull
    public XPathType getExpectedType(XPathExpression expr) {
        return XPathType.UNKNOWN;
    }

    @Nullable
    public QName getQName(PrefixedName qName, XPathElement call) {
        final String prefix = qName.getPrefix();
        final NamespaceContext namespaceContext = getNamespaceContext();
        if (namespaceContext != null) {
            if (prefix != null) {
                final XmlElement element = PsiTreeUtil.getContextOfType(call, XmlElement.class, true);
                final String namespaceURI = namespaceContext.getNamespaceURI(prefix, element);
                return namespaceURI != null && namespaceURI.length() > 0 ? new QName(namespaceURI, qName.getLocalName(), prefix) : null;
            } else {
                return new QName(null, qName.getLocalName(), "");
            }
        } else if (qName.getPrefix() == null) {
            return QName.valueOf(qName.getLocalName());
        } else {
            return null;
        }
    }

    static class DefaultProvider extends ContextProvider {
        private final XmlElement myContextElement;

        DefaultProvider(XmlElement contextElement) {
            myContextElement = contextElement;
        }

        @NotNull
        public ContextType getContextType() {
            return ContextType.PLAIN;
        }

        @Nullable
        public XmlElement getContextElement() {
            return myContextElement;
        }

        @Nullable
        public NamespaceContext getNamespaceContext() {
            return null;
        }

        @Nullable
        public VariableContext getVariableContext() {
            return null;
        }

        @Nullable
        public Set<QName> getAttributes(boolean forValidation) {
            return null;
        }

        @Nullable
        public Set<QName> getElements(boolean forValidation) {
            return null;
        }
    }
}
