/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.associations.impl;

import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.associations.FileAssociationsManager;

import com.intellij.ide.IconUtilEx;
import com.intellij.ide.projectView.ProjectViewNode;
import com.intellij.ide.projectView.impl.AbstractProjectTreeStructure;
import com.intellij.ide.projectView.impl.GroupByTypeComparator;
import com.intellij.ide.projectView.impl.ProjectTreeBuilder;
import com.intellij.ide.util.PsiElementListCellRenderer;
import com.intellij.ide.util.treeView.NodeDescriptor;
import com.intellij.ide.util.treeView.NodeRenderer;
import com.intellij.ide.util.treeView.TreeState;
import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.DimensionService;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiManager;
import com.intellij.uiDesigner.core.GridConstraints;
import com.intellij.util.ui.Tree;

import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.util.*;
import java.util.List;

class AssociationsEditor {
    private static final String KEY = "#XsltSupport.AssociationsEditor";
    private static final Icon LINK_OVERLAY = IconLoader.findIcon("/icons/association_small.png");

    private JPanel myComponent;
    private JPanel myToolbar;
    private JList myList;
    private Tree myTree;
    private JSplitPane mySplitPane;

    private final AssociationsModel myListModel;
    private final TransactionalManager myManager;
    private final ProjectTreeBuilder myBuilder;

    public AssociationsEditor(final Project project, TreeState oldState) {
        myManager = ((FileAssociationsManagerImpl)FileAssociationsManager.getInstance(project)).getTempManager();

        final DefaultActionGroup group = new DefaultActionGroup();
        group.add(new AddAssociationActionWrapper());
        group.add(new RemoveAssociationAction());

        final ActionToolbar toolbar = ActionManager.getInstance().createActionToolbar("AssociationsEditor", group, true);
        myToolbar.add(toolbar.getComponent(), new GridConstraints(0, 0, 1, 1, GridConstraints.ANCHOR_WEST, GridConstraints.FILL_HORIZONTAL, GridConstraints.SIZEPOLICY_CAN_GROW, GridConstraints.SIZEPOLICY_FIXED, null, null, null));

        final DefaultTreeModel treeModel = new DefaultTreeModel(new DefaultMutableTreeNode());
        myBuilder = new ProjectTreeBuilder(project, myTree, treeModel, new MyGroupByTypeComparator(), new MyProjectStructure(project));

        myTree.setModel(treeModel);
        myTree.setCellRenderer(new MyNodeRenderer(myManager));

        if (oldState == null) {
            expandTree(treeModel, project, myBuilder);
        } else {
            oldState.applyTo(myTree);
        }

        myListModel = new AssociationsModel(myTree, myManager);
        myListModel.addListDataListener(new ListDataListener() {
            public void intervalAdded(ListDataEvent listDataEvent) {
                myTree.invalidate();
                myTree.repaint();
            }

            public void intervalRemoved(ListDataEvent listDataEvent) {
                myTree.invalidate();
                myTree.repaint();
            }

            public void contentsChanged(ListDataEvent listDataEvent) {
            }
        });
        myList.setModel(myListModel);
        myList.setCellRenderer(new MyCellRenderer());
        myList.setMinimumSize(new Dimension(120, 200));

        final int location = DimensionService.getInstance().getExtendedState(KEY);
        if (location == -1) {
            myList.setPreferredSize(myList.getMinimumSize());
            mySplitPane.resetToPreferredSizes();
        } else {
            mySplitPane.setDividerLocation(location);
        }
    }

    private void expandTree(DefaultTreeModel newModel, Project project, ProjectTreeBuilder builder) {
        final TreePath rootPath = new TreePath(newModel.getRoot());
        myTree.setSelectionPath(rootPath);

        final PsiManager psiManager = PsiManager.getInstance(project);
        final Set<VirtualFile> files = myManager.getAssociations().keySet();
        if (files.size() > 0) {
            for (VirtualFile file : files) {
                builder.select(psiManager.findFile(file), file, true);
            }
        } else {
            final Enumeration<?> enumeration = ((DefaultMutableTreeNode)myTree.getModel().getRoot()).children();
            while (enumeration.hasMoreElements()) {
                DefaultMutableTreeNode node = (DefaultMutableTreeNode)enumeration.nextElement();
                myTree.expandPath(new TreePath(node.getPath()));
            }
        }
        myTree.setSelectionPath(rootPath);
        myTree.scrollRectToVisible(new Rectangle(new Point(0, 0)));
    }

    public TreeState getState() {
        return TreeState.createOn(myTree);
    }

    public JPanel getComponent() {
        return myComponent;
    }

    @Nullable
    static Object getTreeSelection(JTree tree) {
        final TreePath selectionPath = tree.getSelectionPath();
        if (selectionPath == null) return null;
        final Object component = selectionPath.getLastPathComponent();
        return getObject(component);
    }

    @Nullable
    private static Object getObject(Object component) {
        if (!(component instanceof DefaultMutableTreeNode)) return null;
        final DefaultMutableTreeNode node = ((DefaultMutableTreeNode)component);
        final Object userObject = node.getUserObject();
        if (!(userObject instanceof ProjectViewNode)) return null;
        return ((ProjectViewNode<?>)userObject).getValue();
    }

    public boolean isModified() {
        return myManager.isModified();
    }

    public void apply() {
        myManager.applyChanges();
    }

    public void reset() {
        myManager.reset();
        myListModel.update(null);
    }

    public void dispose() {
        DimensionService.getInstance().setExtendedState(KEY, mySplitPane.getDividerLocation());
        myBuilder.dispose();
        myManager.dispose();
    }

    public void select(PsiFile file) {
        myBuilder.select(file, file.getVirtualFile(), true);
    }

    class AddAssociationActionWrapper extends AddAssociationAction {
        public AddAssociationActionWrapper() {
            super(myManager);
        }

        public void actionPerformed(AnActionEvent e) {
            final PsiFile selection = (PsiFile)getTreeSelection(myTree);
            addAssociation(e, selection);
            myListModel.update(selection);
        }

        public void update(AnActionEvent e) {
            e.getPresentation().setEnabled(getTreeSelection(myTree) instanceof PsiFile);
        }
    }

    class RemoveAssociationAction extends AnAction {
        public RemoveAssociationAction() {
            super(null, "Remove Association", IconLoader.findIcon("/general/remove.png"));
        }

        public void actionPerformed(AnActionEvent e) {
            final PsiFile selection = (PsiFile)getTreeSelection(myTree);
            final PsiFile listSelection = (PsiFile)getListSelection();

            myManager.removeAssociation(selection, listSelection);
            myListModel.update(selection);
        }

        public void update(AnActionEvent e) {
            e.getPresentation().setEnabled(getListSelection() instanceof PsiFile);
        }

        private Object getListSelection() {
            return myList.getSelectedValue();
        }
    }

    private static class MyGroupByTypeComparator extends GroupByTypeComparator {
        protected boolean isSortByType() {
            return false;
        }
    }

    @SuppressWarnings({ "ALL" })
    private static class MyProjectStructure extends AbstractProjectTreeStructure {

        public MyProjectStructure(Project project) {
            super(project);
        }

        public List getProviders() {
            return Collections.EMPTY_LIST;
        }

        public Object[] getChildElements(Object obj) {
            final Object[] childElements = super.getChildElements(obj);
            List l = new ArrayList(childElements.length);
            for (Object o : childElements) {
                if (o instanceof ProjectViewNode) {
                    final ProjectViewNode node = (ProjectViewNode)o;
                    final Object element = node.getValue();
                    if (element instanceof PsiFile) {
                        if (XsltSupport.isXsltFile((PsiFile)element)) {
                            l.add(o);
                        }
                    } else {
                        l.add(o);
                    }
                }
            }
            return l.size() != childElements.length ? (Object[])l.toArray(new Object[l.size()]) : childElements;
        }

        public boolean isShowMembers() {
            return false;
        }

        public boolean isShowModules() {
            return true;
        }

        public boolean isFlattenPackages() {
            return true;
        }

        public boolean isAbbreviatePackageNames() {
            return false;
        }

        public boolean isHideEmptyMiddlePackages() {
            return true;
        }

        public boolean isShowLibraryContents() {
            return false;
        }
    }

    static class AssociationsModel extends AbstractListModel implements TreeSelectionListener {
        private final Tree myTree;
        private final FileAssociationsManager myManager;
        private PsiFile[] myFiles;

        public AssociationsModel(Tree tree, FileAssociationsManager manager) {
            myTree = tree;
            myManager = manager;
            myFiles = PsiFile.EMPTY_ARRAY;
            myTree.addTreeSelectionListener(this);
        }

        public int getSize() {
            return myFiles.length;
        }

        public Object getElementAt(int index) {
            return myFiles[index];
        }

        public void valueChanged(TreeSelectionEvent e) {
            final Object selection = getTreeSelection(myTree);
            if (selection instanceof PsiFile) {
                update((PsiFile)selection);
            } else {
                update(null);
            }
        }

        public void update(PsiFile selection) {
            final int oldSize = myFiles.length;
            myFiles = PsiFile.EMPTY_ARRAY;
            if (myFiles.length != oldSize) fireIntervalRemoved(this, 0, oldSize - 1);
            if (selection != null) {
                myFiles = myManager.getAssociationsFor(selection);
                if (myFiles.length > 0) fireIntervalAdded(this, 0, myFiles.length - 1);
            }
        }
    }

    private static class MyNodeRenderer extends NodeRenderer {
        private final DefaultMutableTreeNode myTemp = new DefaultMutableTreeNode();
        private final FileAssociationsManager myManager;

        public MyNodeRenderer(FileAssociationsManager manager) {
            myManager = manager;
        }

        public void customizeCellRenderer(JTree tree, Object value, boolean selected, boolean expanded, boolean leaf, int row, boolean hasFocus) {
            final Object object = getObject(value);
            if (object instanceof PsiFile) {
                final PsiFile file = (PsiFile)object;
                if (myManager.getAssociationsFor(file).length > 0) {
                    myTemp.setUserObject(new MyNodeDescriptor((NodeDescriptor<?>)((DefaultMutableTreeNode)value).getUserObject()));
                    super.customizeCellRenderer(tree, myTemp, selected, expanded, leaf, row, hasFocus);
                    return;
                }
            }
            super.customizeCellRenderer(tree, value, selected, expanded, leaf, row, hasFocus);
        }

        private class MyNodeDescriptor extends NodeDescriptor<Object> {
            public MyNodeDescriptor(NodeDescriptor<?> nodeDescriptor) {
                super(null, null);
                myName = nodeDescriptor.toString();
                myOpenIcon = nodeDescriptor.getOpenIcon();
                myClosedIcon = IconUtilEx.createLayeredIcon(nodeDescriptor.getClosedIcon(), LINK_OVERLAY);
                myColor = nodeDescriptor.getColor();
            }

            public boolean update() {
                return false;
            }

            public Object getElement() {
                assert false;
                return null;
            }
        }
    }

    private static class MyCellRenderer extends PsiElementListCellRenderer<PsiFile> {
        public String getElementText(PsiFile file) {
            final String name = file.getName();
            assert name != null;
            return name;
        }

        protected String getContainerText(PsiElement psiElement, String string) {
            //noinspection ConstantConditions
            return "(" + ((PsiFile)psiElement).getVirtualFile().getParent().getPresentableUrl() + ")";
        }

        protected int getIconFlags() {
            return 0;
        }
    }
}
