/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.impl;

import org.intellij.lang.xpath.completion.Lookup;
import org.intellij.lang.xpath.completion.XPathCompletionData;
import org.intellij.lang.xpath.xslt.XsltSupport;

import com.intellij.codeInsight.completion.CompletionContext;
import com.intellij.codeInsight.completion.CompletionUtil;
import com.intellij.codeInsight.completion.CompletionVariant;
import com.intellij.codeInsight.completion.XmlCompletionData;
import com.intellij.codeInsight.lookup.LookupItem;
import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.editor.event.EditorFactoryEvent;
import com.intellij.openapi.editor.event.EditorFactoryListener;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiReference;
import com.intellij.psi.filters.ElementFilter;
import com.intellij.psi.xml.XmlAttributeValue;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

import java.util.HashSet;
import java.util.Set;

class XsltCompletionData extends XmlCompletionData implements ApplicationComponent {
    private final XPathCompletionData myXPathCompletionData;
    private final EditorFactoryListener myEditorFactoryListener;

    public XsltCompletionData() {
        final XsltAttributeFilter filter = new XsltAttributeFilter();
        myXPathCompletionData = new XPathCompletionData(filter, XmlAttributeValue.class) {
/*
            protected PsiElement getContext(PsiElement element, int i) {
                PsiElement context = element;

                final XmlAttribute attribute = PsiTreeUtil.getParentOfType(element, XmlAttribute.class);
                if (attribute != null && XsltSupport.isXPathAttribute(attribute)) {
                    final PsiFile[] xpathFiles = XsltSupport.getInstance(element.getProject()).getFiles(attribute);
                    for (PsiFile file : xpathFiles) {
                        final int j = i - element.getTextRange().getStartOffset() - file.getStartOffsetInParent();

                        final PsiElement _context = file.findElementAt(j);
                        if (_context != null) {
                            context = _context;
                            break;
                        }
                    }
                }
                return context;
            }
*/
        };

        final CompletionVariant completionvariant = new MyCompletionVariant(filter);

        completionvariant.includeScopeClass(XmlAttributeValue.class, true);
        completionvariant.addCompletionFilterOnElement(filter);
        completionvariant.setInsertHandler(XPathCompletionData.INSERT_HANDLER);

        registerVariant(completionvariant);

        myEditorFactoryListener = new EditorFactoryListener() {
            public void editorCreated(EditorFactoryEvent event) {
                update(event, XsltCompletionData.this);
            }

            public void editorReleased(EditorFactoryEvent event) {
                update(event, null);
            }

            private void update(EditorFactoryEvent event, XsltCompletionData completionData) {
                if (event.getEditor().isViewer()) return;
                final Project project = event.getEditor().getProject();
                if (project == null) return;

                final PsiDocumentManager dm = PsiDocumentManager.getInstance(project);
                final PsiFile psiFile = dm.getPsiFile(event.getEditor().getDocument());
                if (psiFile != null && XsltSupport.isXsltFile(psiFile)) {
                    //noinspection unchecked
                    psiFile.putUserData(CompletionUtil.ENFORCE_COMPLETION_DATA_KEY, completionData);
                }
            }
        };
    }

    @SuppressWarnings({"RawUseOfParameterizedType"})
    public void addKeywordVariants(Set set, CompletionContext completionContext, PsiElement psiElement) {
        myXPathCompletionData.addKeywordVariants(set, completionContext, psiElement);
        super.addKeywordVariants(set, completionContext, psiElement);
    }

    public String findPrefix(PsiElement psiElement, int i) {
        return myXPathCompletionData.findPrefix(psiElement, i);
    }

    static class MyCompletionVariant extends CompletionVariant {

        public MyCompletionVariant(ElementFilter elementFilter) {
            super(elementFilter);
        }

        @SuppressWarnings({ "unchecked", "RawUseOfParameterizedType" })
        public void addReferenceCompletions(PsiReference psiReference, PsiElement psiElement, Set set, CompletionContext context) {
            final Set s = new HashSet();
            super.addReferenceCompletions(psiReference, psiElement, s, context);
            for (Object o : s) {
                if (o instanceof LookupItem) {
                    final LookupItem item = (LookupItem)o;
                    if (item.getObject() instanceof Lookup) {
                        set.add(item);
                    }
                }
            }
        }

        public boolean hasKeywordCompletions() {
            return false;
        }
    }


    @NotNull
    @NonNls
    public String getComponentName() {
        return "XsltCompletionData";
    }

    public void initComponent() {
        EditorFactory.getInstance().addEditorFactoryListener(myEditorFactoryListener);
    }


    public void disposeComponent() {
        EditorFactory.getInstance().removeEditorFactoryListener(myEditorFactoryListener);
    }
}
