/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.lang.xpath.xslt.impl;

import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.psi.XsltElement;
import org.intellij.lang.xpath.xslt.psi.XsltElementFactory;
import org.intellij.lang.xpath.psi.impl.ResolveUtil;

import com.intellij.navigation.NavigationItem;
import com.intellij.navigation.ItemPresentation;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.progress.ProgressManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.*;
import com.intellij.openapi.startup.StartupManager;
import com.intellij.openapi.vfs.*;
import com.intellij.openapi.editor.colors.TextAttributesKey;
import com.intellij.openapi.vcs.FileStatus;
import com.intellij.psi.*;
import com.intellij.psi.util.CachedValue;
import com.intellij.psi.util.CachedValueProvider;
import com.intellij.psi.xml.XmlFile;
import com.intellij.psi.xml.XmlTag;
import com.intellij.psi.xml.XmlDocument;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.NotNull;

import gnu.trove.THashSet;
import gnu.trove.TObjectHashingStrategy;

import javax.swing.*;
import java.util.*;

/**
 * More or less an adapted copy of com.intellij.lang.javascript.index.JavaScriptIndex
 */
public final class XsltIndex implements ProjectComponent {
    private static final String BUILDING_XSLT_INDEX = "Building XSLT indices";

    private final Project myProject;
    private final PsiManager myPsiManager;

    private Map<XmlFile, XsltIndexEntry> myOldXsltFiles;
    private Map<XmlFile, XsltIndexEntry> myXsltFiles = new HashMap<XmlFile, XsltIndexEntry>();

    private MyTreeChangeListener myTreeChangeListener;
    private VirtualFileListener myFileListener;
    private ModuleRootListener myRootListener;
    private Runnable myUpdateRunnable;

    public XsltIndex(Project project, PsiManager psiManager) {
        myProject = project;
        myPsiManager = psiManager;
    }

    private boolean acceptFile(VirtualFile fileOrDir) {
        if (fileOrDir.getFileType() != StdFileTypes.XML) return false;
        final String extension = fileOrDir.getExtension();
        if (extension == null) return false;
        if (!extension.equalsIgnoreCase("xsl") && !extension.equalsIgnoreCase("xslt")) return false;

        final PsiFile psiFile = myPsiManager.findFile(fileOrDir);
        return psiFile != null && XsltSupport.isXsltFile(psiFile);
    }

    public void projectOpened() {
        myFileListener = new VirtualFileAdapter() {
            @Override
            public void fileCreated(VirtualFileEvent event) {
                fileAdded(event.getFile());
            }

            @Override
            public void beforeFileDeletion(VirtualFileEvent event) {
                final VirtualFile fileOrDir = event.getFile();
                if (acceptFile(fileOrDir)) {
                    processFileRemoved((XmlFile)myPsiManager.findFile(fileOrDir));
                }
            }
        };
        myRootListener = new ModuleRootListener() {
            public void beforeRootsChange(ModuleRootEvent event) {
            }

            public void rootsChanged(ModuleRootEvent event) {
                Runnable runnable = new Runnable() {
                    public void run() {
                        myOldXsltFiles = myXsltFiles;
                        myXsltFiles = new HashMap<XmlFile, XsltIndexEntry>(myOldXsltFiles.size());

                        if (ApplicationManager.getApplication().isUnitTestMode()) {
                            myUpdateRunnable.run();
                        } else {
                            ProgressManager.getInstance().runProcessWithProgressSynchronously(
                                    myUpdateRunnable,
                                    BUILDING_XSLT_INDEX,
                                    false,
                                    myProject
                            );
                        }

                        myOldXsltFiles = null;
                    }
                };
                ApplicationManager.getApplication().invokeLater(runnable);
            }
        };
        myUpdateRunnable = new Runnable() {
            public void run() {
                final ProgressIndicator progress = ProgressManager.getInstance().getProgressIndicator();
                if (progress != null) {
                    progress.setIndeterminate(true);
                    progress.setText(BUILDING_XSLT_INDEX);
                    progress.setText2("");
                }
                final ProjectFileIndex fileIndex = ProjectRootManager.getInstance(myProject).getFileIndex();
                final List<VirtualFile> filesToProcess = new ArrayList<VirtualFile>(5);

                fileIndex.iterateContent(new ContentIterator() {
                    public boolean processFile(VirtualFile fileOrDir) {
                        if (acceptFile(fileOrDir)) {
                            filesToProcess.add(fileOrDir);
                        }
                        return true;
                    }
                });

                if (progress != null) {
                    progress.setIndeterminate(false);
                }

                int processed = 0;
                for (VirtualFile f : filesToProcess) {
                    fileAdded(f);
                    ++processed;
                    if (progress != null) {
                        progress.setFraction((double)processed / filesToProcess.size());
                    }
                }
            }
        };
        myTreeChangeListener = new MyTreeChangeListener();

        if (ApplicationManager.getApplication().isUnitTestMode()) {
            myUpdateRunnable.run();
        } else {
            StartupManager.getInstance(myProject).registerPostStartupActivity(new Runnable() {
                // It seems to be pretty pointless to have all those listeners added before project startup is complete
                // (building indices etc.). Especially the ModuleRootListener seems to be invoked way too often during
                // startup.
                public void run() {
                    Runnable runnable = new Runnable() {
                        public void run() {
                            VirtualFileManager.getInstance().addVirtualFileListener(myFileListener);
                            ProjectRootManager.getInstance(myProject).addModuleRootListener(myRootListener);
                            myPsiManager.addPsiTreeChangeListener(myTreeChangeListener);
                        }
                    };
                    ApplicationManager.getApplication().invokeLater(runnable);

                    myUpdateRunnable.run();
                }
            });
        }
    }

    public void projectClosed() {
        VirtualFileManager.getInstance().removeVirtualFileListener(myFileListener);
        ProjectRootManager.getInstance(myProject).removeModuleRootListener(myRootListener);
        myPsiManager.removePsiTreeChangeListener(myTreeChangeListener);
    }

    @NotNull
    @NonNls
    public String getComponentName() {
        return "XSLT-Support.SymbolIndex";
    }

    public void initComponent() {
    }

    public void disposeComponent() {
        clear();

        myFileListener = null;
        myRootListener = null;
        myUpdateRunnable = null;
        myTreeChangeListener = null;
    }

    public static XsltIndex getInstance(Project project) {
        return project.getComponent(XsltIndex.class);
    }

    private void fileAdded(final VirtualFile fileOrDir) {
        if (!fileOrDir.isDirectory() && acceptFile(fileOrDir)) {
            processFileAdded((XmlFile)myPsiManager.findFile(fileOrDir));
        }
    }

    private void processFileAdded(XmlFile psiFile) {
        if (myOldXsltFiles != null) {
            final XsltIndexEntry indexEntry = myOldXsltFiles.get(psiFile);
            if (indexEntry != null) {
                myXsltFiles.put(psiFile, indexEntry);
                return;
            }
        }

        final ProgressIndicator progress = ProgressManager.getInstance().getProgressIndicator();
        if (progress != null) {
            final VirtualFile virtualFile = psiFile.getVirtualFile();
            progress.setText2(virtualFile != null ? virtualFile.getPresentableUrl() : psiFile.getName());
        }
        myXsltFiles.put(psiFile, new XsltIndexEntry(psiFile));
    }

    private void processFileChanged(XmlFile xmlFile) {
        if (!xmlFile.isPhysical()) return;
        final XsltIndexEntry indexEntry = myXsltFiles.get(xmlFile);
        if (indexEntry == null) {
            processFileAdded(xmlFile);
        }
    }

    private void processFileRemoved(XmlFile xmlFile) {
        myXsltFiles.remove(xmlFile);
    }

    public String[] getSymbolNames() {
        final Set<String> symbolNames = new com.intellij.util.containers.HashSet<String>();
        for (XsltIndexEntry entry : myXsltFiles.values()) {
            entry.fillSymbolNames(symbolNames);
        }
        return symbolNames.toArray(new String[symbolNames.size()]);
    }

    public NavigationItem[] getSymbolsByName(String name) {
        final Set<NavigationItem> symbolNavItems = new com.intellij.util.containers.HashSet<NavigationItem>();
        for (XsltIndexEntry entry : myXsltFiles.values()) {
            entry.fillSymbolsByName(name, symbolNavItems);
        }
        return symbolNavItems.toArray(new NavigationItem[symbolNavItems.size()]);
    }

    public boolean isReachableFrom(XmlFile which, XmlFile from) {
        final XsltIndexEntry target = myXsltFiles.get(which);
        return target != null && _isReachableFrom(which, from);
    }

    private boolean _isReachableFrom(XmlFile which, XmlFile from) {
        final XsltIndexEntry source = myXsltFiles.get(from);
        if (source == null) return false;

        final XsltIndexEntry.IndexEntryContent content = source.myIndexValue.getValue();
        if (content != null) {
            final Set<XmlFile> dependencies = content.myDependencies;
            if (dependencies.contains(which)) {
                return true;
            }
            for (XmlFile dependency : dependencies) {
                if (_isReachableFrom(dependency, which)) {
                    return true;
                }
            }
        }
        return false;
    }

    public void clear() {
        myXsltFiles.clear();
    }

    private class MyTreeChangeListener extends PsiTreeChangeAdapter {
        public void childAdded(PsiTreeChangeEvent event) {
            final PsiElement child = event.getChild();
            if (child instanceof PsiFile && child.isPhysical() && XsltSupport.isXsltFile((PsiFile)child)) {
                processFileAdded((XmlFile)child);
            } else {
                process(event);
            }
        }

        public void childrenChanged(PsiTreeChangeEvent event) {
            process(event);
        }

        public void childRemoved(PsiTreeChangeEvent event) {
            if (event.getChild()instanceof XmlFile) {
                processFileRemoved((XmlFile)event.getChild());
            } else {
                process(event);
            }
        }

        public void childReplaced(PsiTreeChangeEvent event) {
            process(event);
        }

        private void process(PsiTreeChangeEvent event) {
            final PsiElement element = event.getParent();
            if (element != null) {
                final PsiFile psiFile = element.getContainingFile();
                if (psiFile != null && XsltSupport.isXsltFile(psiFile)) {
                    processFileChanged((XmlFile)psiFile);
                }
            }
        }
    }

    static class XsltIndexEntry {

        private final XmlFile myPsiFile;

        static class IndexEntryContent {
            private final Map<String, Object> mySymbols = new HashMap<String, Object>();
            @SuppressWarnings({ "unchecked" })
            private final Set<XmlFile> myDependencies = new THashSet<XmlFile>(TObjectHashingStrategy.IDENTITY);

            void addSymbol(String nameComponent, XsltElement element) {
                final Object o = mySymbols.get(nameComponent);
                if (o == null) {
                    mySymbols.put(nameComponent, element);
                } else if (o instanceof XsltElement) {
                    mySymbols.put(nameComponent, new Object[]{ o, element });
                } else {
                    Object[] oArray = (Object[])o;
                    Object[] newArray = new Object[oArray.length + 1];
                    System.arraycopy(oArray, 0, newArray, 0, oArray.length);
                    newArray[oArray.length] = element;
                    mySymbols.put(nameComponent, newArray);
                }
            }

            public void addReferencedFile(XmlFile ref) {
                myDependencies.add(ref);
            }
        }

        private CachedValue<IndexEntryContent> myIndexValue;

        public XsltIndexEntry(XmlFile psiFile) {
            myPsiFile = psiFile;
            myIndexValue = psiFile.getManager().getCachedValuesManager().createCachedValue(new CachedValueProvider<IndexEntryContent>() {
                public Result<IndexEntryContent> compute() {
                    return new Result<IndexEntryContent>(collectSymbols(), myPsiFile);
                }

                private IndexEntryContent collectSymbols() {
                    final IndexEntryContent content = new IndexEntryContent();

                    final XmlDocument document = myPsiFile.getDocument();
                    if (document == null) {
                        return content;
                    }
                    final XmlTag rootTag = document.getRootTag();
                    if (rootTag == null) {
                        return content;
                    }
                    rootTag.acceptChildren(new PsiElementVisitor() {
                        final XsltElementFactory elementFactory = XsltElementFactory.getInstance();
                        public void visitReferenceExpression(PsiReferenceExpression expression) {
                        }

                        public void visitXmlTag(XmlTag tag) {
                            if (XsltSupport.isVariableOrParam(tag) || XsltSupport.isTemplate(tag)) {
                                content.addSymbol(tag.getAttributeValue("name"), elementFactory.wrapElement(tag));
                            } else if (XsltSupport.isIncludeOrImport(tag)) {
                                final PsiFile ref = ResolveUtil.resolveFile(tag.getAttribute("href", null), myPsiFile);
                                if (ref != null && XsltSupport.isXsltFile(ref)) {
                                    content.addReferencedFile((XmlFile)ref);
                                }
                            }
                        }
                    });
                    return content;
                }
            }, false);

            myIndexValue.getValue();
        }

        public void fillSymbolNames(final Set<String> symbolNames) {
            synchronized (this) {
                final IndexEntryContent content = myIndexValue.getValue();
                if (content != null) symbolNames.addAll(content.mySymbols.keySet());
            }
        }

        public void fillSymbolsByName(final String name, final Set<NavigationItem> symbolNavItems) {
            synchronized (this) {
                if (!myPsiFile.isPhysical() && !ApplicationManager.getApplication().isUnitTestMode()) {
                    return; // predefined
                }

                final IndexEntryContent index = myIndexValue.getValue();
                if (index != null) {
                    final Object value = index.mySymbols.get(name);
                    if (value instanceof Object[]) {
                        for (Object o : (Object[])value) {
                            symbolNavItems.add(new MyNavigationItem((NavigationItem)o));
                        }
                    } else if (value != null) {
                        symbolNavItems.add(new MyNavigationItem((NavigationItem)value));
                    }
                }
            }
        }

        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final XsltIndexEntry that = (XsltIndexEntry)o;

            return myPsiFile == that.myPsiFile;
        }

        public int hashCode() {
            return myPsiFile.hashCode();
        }

        static class MyNavigationItem implements NavigationItem, ItemPresentation {
            private final NavigationItem myItem;
            private ItemPresentation myPresentation;

            public MyNavigationItem(NavigationItem item) {
                myItem = item;
                myPresentation = myItem.getPresentation();
            }

            public String getPresentableText() {
                return myPresentation.getPresentableText();
            }

            @Nullable
            public String getLocationString() {
                return "(in " + ((PsiElement)myItem).getContainingFile().getName() + ")";
            }

            @Nullable
            public Icon getIcon(boolean open) {
                return myPresentation.getIcon(open);
            }

            @Nullable
            public TextAttributesKey getTextAttributesKey() {
                return myPresentation.getTextAttributesKey();
            }

            public String getName() {
                return myItem.getName();
            }

            public ItemPresentation getPresentation() {
                return myPresentation != null ? this : null;
            }

            public FileStatus getFileStatus() {
                return myItem.getFileStatus();
            }

            public void navigate(boolean requestFocus) {
                myItem.navigate(requestFocus);
            }

            public boolean canNavigate() {
                return myItem.canNavigate();
            }

            public boolean canNavigateToSource() {
                return myItem.canNavigateToSource();
            }
        }
    }
}
