/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.lang.xpath.xslt.impl;

import org.intellij.lang.xpath.XPathFile;
import org.intellij.lang.xpath.XPathFileType;
import org.intellij.lang.xpath.XPathTokenTypes;
import org.intellij.lang.xpath.xslt.XsltConfig;
import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.associations.FileAssociationsManager;
import org.intellij.lang.xpath.xslt.context.XsltContextProvider;
import org.intellij.lang.xpath.xslt.impl.references.XsltReferenceProvider;
import org.intellij.lang.xpath.xslt.intentions.AddOptionalArgumentsIntention;
import org.intellij.lang.xpath.xslt.intentions.ConvertIfToChooseIntention;
import org.intellij.lang.xpath.xslt.quickfix.DeleteUnusedParameterFix;
import org.intellij.lang.xpath.xslt.quickfix.DeleteUnusedVariableFix;
import org.intellij.lang.xpath.xslt.util.CompositeDocumentationProvider;

import com.intellij.codeInsight.intention.IntentionManager;
import com.intellij.codeInsight.javadoc.JavaDocManager;
import com.intellij.ide.IconUtilEx;
import com.intellij.ide.projectView.PresentationData;
import com.intellij.ide.projectView.ProjectViewNode;
import com.intellij.ide.projectView.TreeStructureProvider;
import com.intellij.ide.projectView.ViewSettings;
import com.intellij.ide.projectView.impl.nodes.PsiFileNode;
import com.intellij.ide.util.treeView.AbstractTreeNode;
import com.intellij.lang.ParserDefinition;
import com.intellij.lexer.Lexer;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.util.Key;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.util.TextRange;
import com.intellij.psi.InjectedLanguagePlaces;
import com.intellij.psi.LanguageInjector;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiLanguageInjectionHost;
import com.intellij.psi.PsiManager;
import com.intellij.psi.impl.source.resolve.reference.ReferenceProvidersRegistry;
import com.intellij.psi.xml.XmlAttribute;
import com.intellij.psi.xml.XmlAttributeValue;
import com.intellij.psi.xml.XmlElement;
import com.intellij.util.SmartList;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

final class XsltSupportImpl extends XsltSupport implements TreeStructureProvider, LanguageInjector {
    private static final Key<Pair<String, TextRange[]>> CACHED_FILES = Key.create("CACHED_FILES");
    private static final TextRange[] EMPTY_ARRAY = new TextRange[0];

    public static final String[] CATEGORY = { "XSLT" };

    private final XsltConfig myConfig = XsltConfig.getInstance();
    private final Project myProject;
    private final ParserDefinition myParserDefinition;

    XsltSupportImpl(Project project) {
        myProject = project;
        myParserDefinition = XPathFileType.XPATH.getLanguage().getParserDefinition();
    }

    public void projectOpened() {
        if (myConfig.isEnabled()) {
            Runnable runnable = new Runnable() {
                public void run() {
                    final ReferenceProvidersRegistry registry = ReferenceProvidersRegistry.getInstance(myProject);
                    registry.registerReferenceProvider(new XsltAttributeFilter(), XmlAttributeValue.class, new XsltReferenceProvider(myProject));
                }
            };
            ApplicationManager.getApplication().runWriteAction(runnable);

            final JavaDocManager docManager = JavaDocManager.getInstance(myProject);
            final JavaDocManager.DocumentationProvider origProvider = docManager.getProvider(StdFileTypes.XML);
            final XsltDocumentationProvider docProvider = new XsltDocumentationProvider();
            final CompositeDocumentationProvider provider = new CompositeDocumentationProvider(docProvider, origProvider);
            docManager.registerDocumentationProvider(StdFileTypes.XML, provider);
            docManager.registerDocumentationProvider(XPathFileType.XPATH, docProvider);

            final IntentionManager intentionManager = IntentionManager.getInstance(myProject);
            intentionManager.addAction(new DeleteUnusedParameterFix());
            intentionManager.addAction(new DeleteUnusedVariableFix());

            intentionManager.registerIntentionAndMetaData(new AddOptionalArgumentsIntention(), CATEGORY);
            intentionManager.registerIntentionAndMetaData(new ConvertIfToChooseIntention(), CATEGORY);

            PsiManager.getInstance(myProject).registerLanguageInjector(this);
        }
    }

    public void projectClosed() {
    }

    @NotNull
    public String getComponentName() {
        return "XPath-XSLT Integration";
    }

    public void initComponent() {
    }

    public void disposeComponent() {
    }

    public synchronized void getLanguagesToInject(@NotNull PsiLanguageInjectionHost host, @NotNull InjectedLanguagePlaces injectionPlacesRegistrar) {
        if (host instanceof XmlAttributeValue) {
            final PsiElement parent = host.getParent();
            if (parent instanceof XmlAttribute) {
                final XmlAttribute attribute = (XmlAttribute)parent;
                final TextRange[] ranges = getInjectionRanges(attribute);
                if (ranges.length > 0) {
                    XsltContextProvider.createAndAttach((XmlElement)host);
                }
                for (TextRange range : ranges) {
                    // workaround for http://www.jetbrains.net/jira/browse/IDEA-10096
                    if (range instanceof AVTRange) {
                        if (((AVTRange)range).myComplete) {
                            injectionPlacesRegistrar.addPlace(XPathFileType.XPATH.getLanguage(), range.shiftRight(2).grown(-2), "", "");
                        } else {
                            // we need to keep the "'}' expected" parse error
                            injectionPlacesRegistrar.addPlace(XPathFileType.XPATH.getLanguage(), range.shiftRight(2).grown(-1), "{", "");
                        }
                    } else {
                        injectionPlacesRegistrar.addPlace(XPathFileType.XPATH.getLanguage(), range, "", "");
                    }
                }
            }
        }
    }

    @NotNull
    public PsiFile[] getFiles(XmlAttribute attribute) {
        final XmlAttributeValue value = attribute.getValueElement();
        if (value != null) {
            final List<Pair<PsiElement,TextRange>> injectedPsi = ((PsiLanguageInjectionHost)value).getInjectedPsi();
            if (injectedPsi != null) {
                final List<PsiFile> files = new SmartList<PsiFile>();
                for (Pair<PsiElement, TextRange> psi : injectedPsi) {
                    final PsiFile psiFile = (PsiFile)psi.first;
                    if (psiFile instanceof XPathFile) {
                        files.add(psiFile);
                    }
                }
                return files.toArray(new PsiFile[files.size()]);
            }
        }
        return PsiFile.EMPTY_ARRAY;
    }

    @Nullable
    private static TextRange[] getCachedRanges(XmlAttribute attribute) {
        Pair<String, TextRange[]> pair;
        if ((pair = attribute.getUserData(CACHED_FILES)) != null) {
            if (!attribute.getValue().equals(pair.getFirst())) {
                attribute.putUserData(CACHED_FILES, null);
                return null;
            }
        } else {
            return null;
        }
        return pair.getSecond();
    }

    static final class AVTRange extends TextRange {
        final boolean myComplete;

        private AVTRange(int startOffset, int endOffset, boolean iscomplete) {
            super(startOffset, endOffset);
            myComplete = iscomplete;
        }

        public static AVTRange create(XmlAttribute attribute, int startOffset, int endOffset, boolean iscomplete) {
            return new AVTRange(attribute.displayToPhysical(startOffset), attribute.displayToPhysical(endOffset), iscomplete);
        }
    }

    @NotNull
    public TextRange[] getInjectionRanges(final XmlAttribute attribute) {
        if (!XsltSupport.isXsltFile(attribute.getContainingFile())) return EMPTY_ARRAY;
        if (!XsltSupport.isXPathAttribute(attribute)) return EMPTY_ARRAY;

        final TextRange[] cachedFiles = getCachedRanges(attribute);
        if (cachedFiles != null) {
            return cachedFiles;
        }

        final String value = attribute.getDisplayValue();
        if (value == null) return EMPTY_ARRAY;

        final TextRange[] ranges;
        if (mayBeAVT(attribute)) {
            final List<TextRange> avtRanges = new SmartList<TextRange>();

            int i, j = 0;
            final Lexer lexer = myParserDefinition.createLexer(attribute.getProject());
            while ((i = getAVTOffset(value, j)) != -1) {
                // "A right curly brace inside a Literal in an expression is not recognized as terminating the expression."
                lexer.start(value.substring(i).toCharArray());
                j = -1;
                while (lexer.getTokenType() != null) {
                    if (lexer.getTokenType() == XPathTokenTypes.RBRACE) {
                        j = i + lexer.getTokenStart();
                        break;
                    }
                    lexer.advance();
                }

                if (j != -1) {
                    avtRanges.add(AVTRange.create(attribute, i, j + 1, true));
                } else {
                    // missing '}' error will be flagged by xpath parser
                    avtRanges.add(AVTRange.create(attribute, i, value.length(), false));
                    break;
                }
            }

            if (avtRanges.size() > 0) {
                ranges = avtRanges.toArray(new TextRange[avtRanges.size()]);
            } else {
                ranges = EMPTY_ARRAY;
            }
        } else {
            ranges = new TextRange[]{ attribute.getValueTextRange() };
        }

        attribute.putUserData(CACHED_FILES, Pair.create(attribute.getValue(), ranges));

        return ranges;
    }

    @SuppressWarnings({ "RawUseOfParameterizedType", "unchecked" })
    public Collection<AbstractTreeNode> modify(AbstractTreeNode parent, Collection<AbstractTreeNode> children, ViewSettings settings) {
        Collection<AbstractTreeNode> l = children;
        int i=0;
        for (AbstractTreeNode o : children) {
            if (o instanceof ProjectViewNode) {
                final ProjectViewNode node = (ProjectViewNode)o;
                final Object element = node.getValue();
                if (element instanceof PsiFile) {
                    if (XsltSupport.isXsltFile((PsiFile)element)) {
                        if (l == children && l.getClass() != ArrayList.class) {
                            l = new ArrayList<AbstractTreeNode>(children);
                        }
                        final XsltFileNode fileNode = new XsltFileNode(myProject, (PsiFile)element, settings);
                        ((List<AbstractTreeNode>)l).set(i, fileNode);
                    }
                }
            }
            i++;
        }
        return l;
    }

    @Nullable
    @SuppressWarnings({ "RawUseOfParameterizedType" })
    public Object getData(Collection<AbstractTreeNode> selected, String dataName) {
        return null;
    }

    private static class XsltFileNode extends PsiFileNode {
        private static final Icon LINK_OVERLAY = IconLoader.findIcon("/icons/association_small.png");
        private final FileAssociationsManager myInstance;
        private final XsltConfig myConfig;

        public XsltFileNode(Project project, PsiFile psiFile, ViewSettings viewSettings) {
            super(project, psiFile, viewSettings);
            myInstance = FileAssociationsManager.getInstance(myProject);
            myConfig = XsltConfig.getInstance();
        }

        public void updateImpl(PresentationData presentationData) {
            super.updateImpl(presentationData);
            final PsiFile[] psiFiles = myInstance.getAssociationsFor(getValue());

            Icon icon = createXsltIcon(presentationData.getIcon(false));
            if (psiFiles.length > 0) {
                icon = IconUtilEx.createLayeredIcon(icon, LINK_OVERLAY);
            }
            presentationData.setIcons(icon);
        }

        public boolean expandOnDoubleClick() {
            return false;
        }

        @SuppressWarnings({ "RawUseOfParameterizedType" })
        public Collection<AbstractTreeNode> getChildrenImpl() {
            if (myConfig.isShowLinkedFiles()) {
                final PsiFile[] psiFiles = myInstance.getAssociationsFor(getValue());
                if (psiFiles.length > 0) {
                    return ProjectViewNode.wrap(Arrays.asList(psiFiles), myProject, PsiFileNode.class, getSettings());
                }
            }
            return super.getChildrenImpl();
        }
    }
}
