/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.psi.impl;

import org.intellij.lang.xpath.XPathFileType;
import org.intellij.lang.xpath.psi.XPathVariable;
import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.psi.XsltParameter;
import org.intellij.lang.xpath.xslt.psi.XsltTemplate;

import com.intellij.formatting.FormattingModelBuilder;
import com.intellij.ide.structureView.StructureViewBuilder;
import com.intellij.lang.Commenter;
import com.intellij.lang.Language;
import com.intellij.lang.PairedBraceMatcher;
import com.intellij.lang.ParserDefinition;
import com.intellij.lang.annotation.Annotator;
import com.intellij.lang.annotation.ExternalAnnotator;
import com.intellij.lang.cacheBuilder.WordsScanner;
import com.intellij.lang.findUsages.FindUsagesProvider;
import com.intellij.lang.folding.FoldingBuilder;
import com.intellij.lang.refactoring.NamesValidator;
import com.intellij.lang.refactoring.RefactoringSupportProvider;
import com.intellij.lang.surroundWith.SurroundDescriptor;
import com.intellij.lang.xml.XMLLanguage;
import com.intellij.navigation.ItemPresentation;
import com.intellij.navigation.NavigationItem;
import com.intellij.openapi.fileTypes.SyntaxHighlighter;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiNamedElement;
import com.intellij.psi.tree.TokenSet;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlTag;
import com.intellij.refactoring.RefactoringActionHandler;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class XsltLanguage extends Language {
    public static final String ID = "$XSLT";
    public static final XsltLanguage INSTANCE = new XsltLanguage();

    private final Language XML = Language.findInstance(XMLLanguage.class);
    private final FindUsagesProvider FIND_USAGES_PROVIDER = new MyFindUsagesProvider();

    XsltLanguage() {
        super(ID);
    }

    @Override
    @NotNull
    public FindUsagesProvider getFindUsagesProvider() {
        return FIND_USAGES_PROVIDER;
    }

    @Override
    @NotNull
    public NamesValidator getNamesValidator() {
        return XPathFileType.XPATH.getLanguage().getNamesValidator();
    }

    @Override
    @NotNull
    public RefactoringSupportProvider getRefactoringSupportProvider() {
        return new RefactoringSupportProvider() {
            public boolean isSafeDeleteAvailable(PsiElement element) {
                if (element instanceof XPathVariable) {
                    return true;
                }
                return element instanceof XsltTemplate;
            }

            @Nullable
            public RefactoringActionHandler getIntroduceVariableHandler() {
                return null;
            }
        };
    }

    @Override
    @NotNull
    public SyntaxHighlighter getSyntaxHighlighter(Project project, VirtualFile file) {
        return XML.getSyntaxHighlighter(project, file);
    }

    @Override
    public FormattingModelBuilder getFormattingModelBuilder() {
        return XML.getFormattingModelBuilder();
    }

    @Override
    public ParserDefinition getParserDefinition() {
        return XML.getParserDefinition();
    }

    @Override
    public FoldingBuilder getFoldingBuilder() {
        return XML.getFoldingBuilder();
    }

    @Override
    public PairedBraceMatcher getPairedBraceMatcher() {
        return XML.getPairedBraceMatcher();
    }

    @Override
    public Commenter getCommenter() {
        return XML.getCommenter();
    }

    @Override
    @NotNull
    public TokenSet getReadableTextContainerElements() {
        return XML.getReadableTextContainerElements();
    }

    @Override
    public Annotator getAnnotator() {
        return XML.getAnnotator();
    }

    @Override
    public ExternalAnnotator getExternalAnnotator() {
        return XML.getExternalAnnotator();
    }

    @Override
    public StructureViewBuilder getStructureViewBuilder(PsiFile psiFile) {
        return XML.getStructureViewBuilder(psiFile);
    }

    @Override
    @NotNull
    public SurroundDescriptor[] getSurroundDescriptors() {
        return XML.getSurroundDescriptors();
    }

    @Override
    public String[] getMimeTypes() {
        return XML.getMimeTypes();
    }

    private class MyFindUsagesProvider implements FindUsagesProvider {
        @Nullable
        public WordsScanner getWordsScanner() {
            return XML.getFindUsagesProvider().getWordsScanner();
        }

        public boolean canFindUsagesFor(PsiElement psiElement) {
            return psiElement instanceof PsiNamedElement;
        }

        @Nullable
        public String getHelpId(PsiElement psiElement) {
            return null;
        }

        @NotNull
        public String getType(PsiElement element) {
            if (element instanceof XsltParameter) {
                return getParameterType((XsltParameter)element);
            }
            if (element instanceof XPathVariable) return "variable";
            if (element instanceof XsltTemplate) return "template";
            if (element instanceof ImplicitModeElement) return "mode";
            return "";
        }

        private String getParameterType(XsltParameter myTarget) {
            final XmlTag parentTag = PsiTreeUtil.getParentOfType(myTarget.getNavigationElement(), XmlTag.class);
            if (parentTag != null) {
                if (XsltSupport.isXsltRootTag(parentTag)) {
                    return "stylesheet parameter";
                } else if (XsltSupport.isTemplate(parentTag, false)) {
                    return "template parameter";
                }
            }
            return "parameter";
        }

        @NotNull
        public String getDescriptiveName(PsiElement element) {
            if (element instanceof PsiNamedElement) {
                final String name = ((PsiNamedElement)element).getName();
                if (name != null) return name;
            }
            return element.toString();
        }

        @NotNull
        public String getNodeText(PsiElement element, boolean useFullName) {
            if (useFullName) {
                if (element instanceof NavigationItem) {
                    final NavigationItem navigationItem = ((NavigationItem)element);
                    final ItemPresentation presentation = navigationItem.getPresentation();
                    if (presentation != null && presentation.getPresentableText() != null) {
                        return presentation.getPresentableText();
                    }
                    final String name = navigationItem.getName();
                    if (name != null) {
                        return name;
                    }
                }
            }
            if (element instanceof PsiNamedElement) {
                final String name = ((PsiNamedElement)element).getName();
                if (name != null) return name;
            }
            return element.toString();
        }
    }
}
