/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.quickfix;

import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.util.XsltCodeInsightUtil;

import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiNamedElement;
import com.intellij.psi.PsiReference;
import com.intellij.psi.search.LocalSearchScope;
import com.intellij.psi.search.searches.ReferencesSearch;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlTag;
import com.intellij.util.IncorrectOperationException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;

public abstract class DeleteUnusedElementBase extends AbstractFix {
    private String myName;

    @NotNull
    public String getText() {
        return "Delete unused " + getType() + " '" + myName + "'";
    }

    public boolean isAvailable(Project project, Editor editor, PsiFile file) {
        if (!XsltSupport.isXsltFile(file)) return false;

        final PsiElement element = file.findElementAt(editor.getCaretModel().getOffset());
        if (element != null) {
            final XmlTag tag = PsiTreeUtil.getParentOfType(element, XmlTag.class, false);
            final PsiNamedElement obj = getElement(tag);
            if (obj != null && obj.isValid()) {
                myName = obj.getName();
                if (myName != null && myName.length() > 0) {
                    final XmlTag templateTag = XsltCodeInsightUtil.getTemplateTag(tag, false);
                    if (templateTag != null) {
                        final LocalSearchScope searchScope = new LocalSearchScope(templateTag);
                        return isUnused(obj, ReferencesSearch.search(obj, searchScope, false).findAll());
                    }
                }
            }
        }
        return false;
    }

    public void invoke(Project project, Editor editor, PsiFile file) throws IncorrectOperationException {
        assert isAvailable(project, editor, file);

        final XmlTag tag = PsiTreeUtil.getParentOfType(file.findElementAt(editor.getCaretModel().getOffset()), XmlTag.class, false);
        assert tag != null;

        final PsiElement obj = getElement(tag);
        assert obj != null;

        deleteElement(file, obj);
    }

    protected abstract String getType();

    @Nullable
    protected abstract PsiNamedElement getElement(@Nullable XmlTag tag);

    protected void deleteElement(PsiFile file, @NotNull PsiElement obj) throws IncorrectOperationException {
        obj.delete();
    }

    protected boolean isUnused(PsiElement obj, Collection<PsiReference> references) {
        return references.size() == 0;
    }
}