/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.quickfix;

import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.util.XsltCodeInsightUtil;
import org.intellij.lang.xpath.xslt.psi.XsltElement;
import org.intellij.lang.xpath.xslt.psi.XsltElementFactory;
import org.intellij.lang.xpath.xslt.psi.XsltParameter;
import org.intellij.lang.xpath.xslt.psi.XsltTemplate;

import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiNamedElement;
import com.intellij.psi.PsiReference;
import com.intellij.psi.search.SearchScope;
import com.intellij.psi.search.searches.ReferencesSearch;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlAttribute;
import com.intellij.psi.xml.XmlAttributeValue;
import com.intellij.psi.xml.XmlTag;
import com.intellij.util.IncorrectOperationException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;

public class DeleteUnusedParameterFix extends DeleteUnusedElementBase {

    protected String getType() {
        return "Parameter";
    }

    @Nullable
    protected PsiNamedElement getElement(@Nullable XmlTag tag) {
        if (tag != null) {
            final XmlAttribute attribute = tag.getAttribute("name", null);
            if (attribute != null && XsltSupport.isParam(attribute)) {
                //noinspection RedundantCast
                return (PsiNamedElement)(XsltElement)XsltElementFactory.getInstance().wrapElement(tag);
            }
        }
        return null;
    }

    protected void deleteElement(PsiFile file, @NotNull PsiElement obj) throws IncorrectOperationException {
        final XsltTemplate template = XsltCodeInsightUtil.getTemplate(((XsltParameter)obj).getTag(), false);
        if (template == null || template.getMatchExpression() == null) {
            final SearchScope searchScope = obj.getResolveScope();
            final Collection<PsiReference> references = ReferencesSearch.search(obj, searchScope, false).findAll();
            for (PsiReference reference : references) {
                final XmlTag t = PsiTreeUtil.getContextOfType(reference.getElement(), XmlTag.class, true);
                if (t != null && XsltSupport.XSLT_NS.equals(t.getNamespace())) {
                    assert "with-param".equals(t.getLocalName());
                    t.delete();
                }
            }
        }
        super.deleteElement(file, obj);
    }

    protected boolean isUnused(PsiElement obj, Collection<PsiReference> references) {
        int n = references.size();
        for (PsiReference reference : references) {
            final PsiElement element = reference.getElement();
            if (element instanceof XmlAttributeValue) {
                final XmlAttribute parent = (XmlAttribute)element.getParent();
                if ("name".equals(parent.getName())) {
                    final XmlTag tag = parent.getParent();
                    if (tag != null && "with-param".equals(tag.getLocalName())) {
                        n--;
                    }
                }
            }
        }
        return n == 0;
    }
}
