/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.run;

import static org.intellij.lang.xpath.xslt.run.XsltRunConfiguration.isEmpty;

import com.intellij.execution.CantRunException;
import com.intellij.execution.ExecutionException;
import com.intellij.execution.configurations.*;
import com.intellij.execution.process.OSProcessHandler;
import com.intellij.execution.process.ProcessAdapter;
import com.intellij.execution.process.ProcessEvent;
import com.intellij.ide.BrowserUtil;
import com.intellij.ide.plugins.PluginManager;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.extensions.PluginId;
import com.intellij.openapi.fileEditor.OpenFileDescriptor;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.projectRoots.ProjectJdk;
import com.intellij.openapi.roots.ProjectRootsTraversing;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VirtualFileManager;

import javax.swing.*;
import java.io.File;

class XsltCommandLineState extends JavaCommandLineState {
    private final XsltRunConfiguration myXsltRunConfiguration;

    @SuppressWarnings({ "RawUseOfParameterizedType" })
    public XsltCommandLineState(XsltRunConfiguration xsltRunConfiguration, RunnerSettings runnerSettings, ConfigurationPerRunnerSettings configurationSettings) {
        super(runnerSettings, configurationSettings);
        myXsltRunConfiguration = xsltRunConfiguration;
    }

    protected OSProcessHandler startProcess() throws ExecutionException {
        final OSProcessHandler osProcessHandler = super.startProcess();
        osProcessHandler.addProcessListener(new MyProcessAdapter());
        return osProcessHandler;
    }

    protected JavaParameters createJavaParameters() throws ExecutionException {
        final ProjectJdk jdk = myXsltRunConfiguration.getEffectiveJDK();
        if (jdk == null) {
            throw CantRunException.noJdkConfigured();
        }
        final JavaParameters parameters = new JavaParameters();
        parameters.setJdk(jdk);

        if (myXsltRunConfiguration.getJdkChoice() == XsltRunConfiguration.JdkChoice.FROM_MODULE) {
            final Module module = myXsltRunConfiguration.getEffectiveModule();
            if (module == null) {
                throw CantRunException.noModuleConfigured("");
            }
            ProjectRootsTraversing.collectRoots(module, ProjectRootsTraversing.FULL_CLASSPATH_WITHOUT_TESTS, parameters.getClassPath());
        }

        final ParametersList vmParameters = parameters.getVMParametersList();
        vmParameters.addParametersString(myXsltRunConfiguration.myVmArguments);
        if (isEmpty(myXsltRunConfiguration.getXsltFile())) {
            throw new CantRunException("No XSLT file selected");
        }
        vmParameters.defineProperty("xslt.file", myXsltRunConfiguration.getXsltFile());
        if (isEmpty(myXsltRunConfiguration.getXmlInputFile())) {
            throw new CantRunException("No XML input file selected");
        }
        vmParameters.defineProperty("xslt.input", myXsltRunConfiguration.getXmlInputFile());

        final XsltRunConfiguration.OutputType outputType = myXsltRunConfiguration.getOutputType();
        if (outputType == XsltRunConfiguration.OutputType.CONSOLE) {
            vmParameters.defineProperty("xslt.listen-port", String.valueOf(myXsltRunConfiguration.myRunnerPort));
        } else if (outputType == XsltRunConfiguration.OutputType.FILE) {
            vmParameters.defineProperty("xslt.output",  myXsltRunConfiguration.myOutputFile);
        }

        for (Pair<String, String> pair : myXsltRunConfiguration.getParameters()) {
            final String name = pair.getFirst();
            final String value = pair.getSecond();
            if (isEmpty(name) || value == null) continue;
            vmParameters.defineProperty("xslt.param." + name, value);
        }
        vmParameters.defineProperty("xslt.smart-error-handling", String.valueOf(myXsltRunConfiguration.mySmartErrorHandling));

        final PluginId pluginId = PluginManager.getPluginByClassName(getClass().getName());
        final File pluginPath = PluginManager.getPlugin(pluginId).getPath();
        final char c = File.separatorChar;
        File rtClasspath = new File(pluginPath, "lib" + c + "rt" + c + "xslt-rt.jar");
        if (!rtClasspath.exists()) {
            if (!(rtClasspath = new File(pluginPath, "classes")).exists()) {
                final CantRunException ex = new CantRunException("Runtime classes not found");
                Logger.getInstance(getClass().getName()).error(ex);
                throw ex;
            }
        }
        parameters.getClassPath().addTail(rtClasspath.getAbsolutePath());

        parameters.setMainClass("org.intellij.lang.xpath.xslt.run.rt.XSLTRunner");

        parameters.setWorkingDirectory(isEmpty(myXsltRunConfiguration.myWorkingDirectory) ?
                new File(myXsltRunConfiguration.getXsltFile()).getParentFile().getAbsolutePath() :
                myXsltRunConfiguration.myWorkingDirectory);

        return parameters;
    }

    private class MyProcessAdapter extends ProcessAdapter {

        public void processTerminated(final ProcessEvent event) {

            if (myXsltRunConfiguration.getOutputType() == XsltRunConfiguration.OutputType.FILE) {
                Runnable runnable = new Runnable() {
                    public void run() {
                        Runnable runnable = new Runnable() {
                            public void run() {
                                if (event.getExitCode() == 0) {
                                    if (myXsltRunConfiguration.myOpenInBrowser) {
                                        BrowserUtil.launchBrowser(myXsltRunConfiguration.myOutputFile);
                                    }
                                    if (myXsltRunConfiguration.myOpenOutputFile) {
                                        final String url = VfsUtil.pathToUrl(myXsltRunConfiguration.myOutputFile);
                                        final VirtualFile fileByUrl = VirtualFileManager.getInstance().refreshAndFindFileByUrl(url.replace(File.separatorChar, '/'));
                                        if (fileByUrl != null) {
                                            fileByUrl.refresh(false, false);
                                            new OpenFileDescriptor(myXsltRunConfiguration.getProject(), fileByUrl).navigate(true);
                                            return;
                                        }
                                    }
                                    VirtualFileManager.getInstance().refresh(true);
                                }
                            }
                        };
                        ApplicationManager.getApplication().runWriteAction(runnable);
                    }
                };
                SwingUtilities.invokeLater(runnable);
            }
        }
    }
}