/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.run;

import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.associations.FileAssociationsManager;

import com.intellij.execution.ExecutionException;
import com.intellij.execution.configurations.*;
import com.intellij.execution.filters.RegexpFilter;
import com.intellij.execution.filters.TextConsoleBuilder;
import com.intellij.execution.filters.TextConsoleBuilderFactory;
import com.intellij.execution.process.*;
import com.intellij.execution.runners.JavaProgramRunner;
import com.intellij.execution.runners.RunnerInfo;
import com.intellij.openapi.actionSystem.DataContext;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.openapi.fileTypes.FileTypeManager;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.module.ModuleManager;
import com.intellij.openapi.options.SettingsEditor;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.projectRoots.ProjectJdk;
import com.intellij.openapi.projectRoots.ProjectJdkTable;
import com.intellij.openapi.roots.ModuleRootManager;
import com.intellij.openapi.roots.ProjectRootManager;
import com.intellij.openapi.util.*;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.pointers.VirtualFilePointer;
import com.intellij.openapi.vfs.pointers.VirtualFilePointerManager;
import com.intellij.psi.PsiFile;
import com.intellij.psi.xml.XmlDocument;
import com.intellij.psi.xml.XmlFile;
import com.intellij.psi.xml.XmlTag;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import org.jdom.Element;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

class XsltRunConfiguration extends RunConfigurationBase implements LocatableConfiguration {
    private static final String NAME = "XSLT Configuration";
    static final String XSLT_TAB_NAME = "XSLT Output";

    private static final String STRICT_FILE_PATH_EXPR = "(file\\://?(?:/?\\p{Alpha}\\:)?(?:/\\p{Alpha}\\:)?[^:]+)";
    private static final String RELAXED_FILE_PATH_EXPR = "((?:file\\://?)?(?:/?\\p{Alpha}\\:)?(?:/\\p{Alpha}\\:)?[^:]+)";

    private static final String LOG_TAG = "(?:\\[[\\w ]+\\]\\:? +)?";
    private static final VirtualFilePointerManager FILE_POINTER_MANAGER = VirtualFilePointerManager.getInstance();

    public enum OutputType {
        CONSOLE, STDOUT, FILE
    }

    public enum JdkChoice {
        FROM_MODULE, JDK
    }

    private List<Pair<String, String>> myParameters = new ArrayList<Pair<String, String>>();
    private @Nullable VirtualFilePointer myXsltFile = null;
    private @Nullable VirtualFilePointer myXmlInputFile = null;
    private @NotNull OutputType myOutputType = OutputType.CONSOLE;
    private @NotNull JdkChoice myJdkChoice = JdkChoice.FROM_MODULE;
    private @Nullable FileType myFileType = StdFileTypes.XML;

    public String myOutputFile; // intentionally untracked. should it be?
    public boolean myOpenOutputFile;
    public boolean myOpenInBrowser;
    public boolean mySmartErrorHandling = true;
    public int myRunnerPort = 34873;
    public String myVmArguments;
    public String myWorkingDirectory;
    public String myModule;
    public String myJdk;

    private final String mySuggestedName;

    private XsltRunConfiguration(Project project, ConfigurationFactory factory, String suggestedName) {
        super(project, factory, NAME);
        mySuggestedName = suggestedName;
    }

    public XsltRunConfiguration(Project project, ConfigurationFactory factory) {
        super(project, factory, NAME);
        mySuggestedName = null;
    }

    public SettingsEditor<XsltRunConfiguration> getConfigurationEditor() {
        return new XsltRunSettingsEditor(getProject());
    }

    @SuppressWarnings({"ConstantConditions"})
    public JDOMExternalizable createRunnerSettings(ConfigurationInfoProvider provider) {
        return null;  // nothing special here
    }

    @SuppressWarnings({ "ConstantConditions", "RawUseOfParameterizedType" })
    public SettingsEditor<JDOMExternalizable> getRunnerSettingsEditor(JavaProgramRunner runner) {
        return null;  // nothing special here
    }

    @SuppressWarnings({ "RawUseOfParameterizedType" })
    public RunProfileState getState(DataContext context,
                                    RunnerInfo runnerInfo,
                                    RunnerSettings runnerSettings,
                                    ConfigurationPerRunnerSettings configurationSettings) throws ExecutionException {

        if (myXsltFile == null) throw new ExecutionException("No XSLT file selected");
        final VirtualFile baseFile = myXsltFile.getFile();

        final XsltCommandLineState state = new XsltCommandLineState(this, runnerSettings, configurationSettings);
        final TextConsoleBuilder builder = TextConsoleBuilderFactory.getInstance().createBuilder(getProject());
        builder.addFilter(new CustomRegexpFilter(getProject(), RegexpFilter.FILE_PATH_MACROS + "\\:" +
                "(?:(?: line )?" + RegexpFilter.LINE_MACROS + ")?" +
                "(?:\\:(?: column )?" + RegexpFilter.COLUMN_MACROS + ")?", baseFile, STRICT_FILE_PATH_EXPR));
        builder.addFilter(new CustomRegexpFilter(getProject(), LOG_TAG +
                RegexpFilter.FILE_PATH_MACROS + "\\:" +
                "(?:(?: line )?" + RegexpFilter.LINE_MACROS + ")?" +
                "(?:\\:(?: column )?" + RegexpFilter.COLUMN_MACROS + ")?", baseFile, RELAXED_FILE_PATH_EXPR));

        builder.addFilter(new CustomRegexpFilter(getProject(), RegexpFilter.FILE_PATH_MACROS + ";" +
                " \\w+ #" + RegexpFilter.LINE_MACROS +
                "(?:; \\w+ #" + RegexpFilter.COLUMN_MACROS + ")?", baseFile, STRICT_FILE_PATH_EXPR));
        builder.addFilter(new CustomRegexpFilter(getProject(), LOG_TAG +
                RegexpFilter.FILE_PATH_MACROS + ";" +
                " \\w+ #" + RegexpFilter.LINE_MACROS +
                "(?:; \\w+ #" + RegexpFilter.COLUMN_MACROS + ")?", baseFile, RELAXED_FILE_PATH_EXPR));

        builder.addFilter(new CustomRegexpFilter(getProject(), "(?:" + RegexpFilter.FILE_PATH_MACROS + ")?" +
                " line " + RegexpFilter.LINE_MACROS +
                "(?:\\:(?: column )?" + RegexpFilter.COLUMN_MACROS + ")?", baseFile, STRICT_FILE_PATH_EXPR));

        state.setConsoleBuilder(builder);
        return state;
    }

    //invoke before run/debug tabs are shown.
    //Should be overriden to add additional tabs for run/debug toolwindow
    @Override
    public void createAdditionalTabComponents(AdditionalTabComponentManager manager, ProcessHandler startedProcess) {
        if (myOutputType == OutputType.CONSOLE && myFileType != null) {
            final HighlightingOutputConsole console = new HighlightingOutputConsole(getProject(), myFileType);

            final OutputTabAdapter listener = new OutputTabAdapter(this, startedProcess, console);
            if (startedProcess.isStartNotified()) {
                listener.attachOutputConsole();
            } else {
                startedProcess.addProcessListener(listener);
            }
            manager.addAdditionalTabComponent(console);
        }
    }

    public final RunConfiguration clone() {
        final XsltRunConfiguration configuration = (XsltRunConfiguration)super.clone();
        configuration.myParameters = new ArrayList<Pair<String, String>>(myParameters);
        if (myXsltFile != null) configuration.myXsltFile = FILE_POINTER_MANAGER.duplicate(myXsltFile, null);
        if (myXmlInputFile != null) configuration.myXmlInputFile = FILE_POINTER_MANAGER.duplicate(myXmlInputFile, null);
        return configuration;
    }

    public void checkConfiguration() throws RuntimeConfigurationException {
        if (myXsltFile == null) {
            throw new RuntimeConfigurationError("No XSLT File selected");
        }
        if (myXsltFile.getFile() == null) {
            throw new RuntimeConfigurationError("Selected XSLT File not found");
        }
        if (myXmlInputFile == null) {
            throw new RuntimeConfigurationError("No XML Input File selected");
        }
        if (myXmlInputFile.getFile() == null) {
            throw new RuntimeConfigurationError("Selected XML Input File not found");
        }
        if (myOutputType == OutputType.FILE) {
            if (isEmpty(myOutputFile)) {
                throw new RuntimeConfigurationError("No output file selected");
            }
            final File f = new File(myOutputFile);
            if (f.isDirectory()) {
                throw new RuntimeConfigurationError("Selected output file points to a directory");
            } else if (!f.canWrite()) {
                throw new RuntimeConfigurationError("Selected output file is not writable");
            }
        }
        if (getEffectiveJDK() == null) {
            throw new RuntimeConfigurationError("No JDK available");
        }
    }

    static boolean isEmpty(String file) {
        return file == null || file.length() == 0;
    }

    // return modules to compile before run. Null or empty list to make project
    public Module[] getModules() {
        return getModule() != null ? new Module[]{ getModule() } : Module.EMPTY_ARRAY;
    }

    @SuppressWarnings({ "unchecked" })
    public void readExternal(Element element) throws InvalidDataException {
        super.readExternal(element);
        DefaultJDOMExternalizer.readExternal(this, element);

        Element e = element.getChild("XsltFile");
        if (e != null) {
            myXsltFile = FILE_POINTER_MANAGER.create("", null);
            assert myXsltFile != null;
            myXsltFile.readExternal(e);
        }
        e = element.getChild("XmlFile");
        if (e != null) {
            myXmlInputFile = FILE_POINTER_MANAGER.create("", null);
            assert myXmlInputFile != null;
            myXmlInputFile.readExternal(e);
        }

        final Element parameters = element.getChild("parameters");
        if (parameters != null) {
            myParameters.clear();
            final List<Element> params = parameters.getChildren("param");
            for (Element p : params) {
                myParameters.add(Pair.create(p.getAttributeValue("name"), p.getAttributeValue("value")));
            }
        }

        final Element outputType = element.getChild("OutputType");
        if (outputType != null) {
            myOutputType = OutputType.valueOf(outputType.getAttributeValue("value"));
        }
        final Element fileType = element.getChild("FileType");
        if (fileType != null) {
            myFileType = getFileType(fileType.getAttributeValue("name"));
        }
        final Element jdkChoice = element.getChild("JdkChoice");
        if (jdkChoice != null) {
            myJdkChoice = JdkChoice.valueOf(jdkChoice.getAttributeValue("value"));
        }
    }

    @Nullable
    private FileType getFileType(String value) {
        final FileType[] fileTypes = FileTypeManager.getInstance().getRegisteredFileTypes();
        for (FileType fileType : fileTypes) {
            if (fileType.getName().equals(value)) {
                return fileType;
            }
        }
        return null;
    }

    public void writeExternal(Element element) throws WriteExternalException {
        super.writeExternal(element);
        DefaultJDOMExternalizer.writeExternal(this, element);

        Element e = new Element("XsltFile");
        if (myXsltFile != null) {
            myXsltFile.writeExternal(e);
            element.addContent(e);
        }
        e = new Element("XmlFile");
        if (myXmlInputFile != null) {
            myXmlInputFile.writeExternal(e);
            element.addContent(e);
        }

        final Element params = new Element("parameters");
        element.addContent(params);
        for (Pair<String, String> pair : myParameters) {
            final Element p = new Element("param");
            params.addContent(p);
            p.setAttribute("name", pair.getFirst());

            final String value = pair.getSecond();
            if (value != null) {
                p.setAttribute("value", value);
            }
        }

        final Element type = new Element("OutputType");
        type.setAttribute("value", myOutputType.name());
        element.addContent(type);

        if (myFileType != null) {
            final String name = myFileType.getName();
            final Element fileType = new Element("FileType");
            fileType.setAttribute("name", name);
            element.addContent(fileType);
        }

        final Element choice = new Element("JdkChoice");
        choice.setAttribute("value", myJdkChoice.name());
        element.addContent(choice);
    }

    public List<Pair<String, String>> getParameters() {
        return myParameters;
    }

    public void setParameters(List<Pair<String, String>> params) {
        myParameters.clear();
        myParameters.addAll(params);
    }

    public String getVmArguments() {
        return myVmArguments;
    }

    public void setVmArguments(String vmArguments) {
        myVmArguments = vmArguments;
    }

    public void setXsltFile(String xsltFile) {
        if (myXsltFile != null) {
            FILE_POINTER_MANAGER.kill(myXsltFile);
        }
        if (isEmpty(xsltFile)) {
            myXsltFile = null;
        } else {
            myXsltFile = FILE_POINTER_MANAGER.create(VfsUtil.pathToUrl(xsltFile).replace(File.separatorChar, '/'), null);
        }
    }
    private void setXsltFile(VirtualFile virtualFile) {
        if (myXsltFile != null) {
            FILE_POINTER_MANAGER.kill(myXsltFile);
        }
        myXsltFile = FILE_POINTER_MANAGER.create(virtualFile, null);
    }

    @Nullable
    public String getXsltFile() {
        return myXsltFile != null ? myXsltFile.getPresentableUrl() : null;
    }

    @Nullable
    public String getXmlInputFile() {
        return myXmlInputFile != null ? myXmlInputFile.getPresentableUrl() : null;
    }

    @Nullable
    public FileType getFileType() {
        return myFileType;
    }

    public void setFileType(FileType fileType) {
        myFileType = fileType;
    }

    @Nullable
    public Module getModule() {
        return myModule != null ? ModuleManager.getInstance(getProject()).findModuleByName(myModule) : null;
    }

    @Nullable
    public ProjectJdk getJdk() {
        return myJdk != null ? ProjectJdkTable.getInstance().findJdk(myJdk) : null;
    }

    public void setXmlInputFile(String xmlInputFile) {
        if (myXmlInputFile != null) {
            FILE_POINTER_MANAGER.kill(myXmlInputFile);
        }
        if (isEmpty(xmlInputFile)) {
            myXmlInputFile = null;
        } else {
            myXmlInputFile = FILE_POINTER_MANAGER.create(VfsUtil.pathToUrl(xmlInputFile).replace(File.separatorChar, '/'), null);
        }
    }

    public void setXmlInputFile(VirtualFile xmlInputFile) {
        FILE_POINTER_MANAGER.kill(myXmlInputFile);
        myXmlInputFile = FILE_POINTER_MANAGER.create(xmlInputFile, null);
    }

    public void setModule(Module module) {
        myModule = module != null ? module.getName() : null;
    }

    public void setJDK(ProjectJdk projectJdk) {
        myJdk = projectJdk != null ? projectJdk.getName() : null;
    }

    public OutputType getOutputType() {
        return myOutputType;
    }

    public void setOutputType(OutputType outputType) {
        myOutputType = outputType;
    }

    public JdkChoice getJdkChoice() {
        return myJdkChoice;
    }

    public void setJdkChoice(JdkChoice jdkChoice) {
        myJdkChoice = jdkChoice;
    }

    @Nullable
    ProjectJdk getEffectiveJDK() {
        if (myJdkChoice == JdkChoice.JDK) {
            return myJdk != null ? ProjectJdkTable.getInstance().findJdk(myJdk) : null;
        }
        ProjectJdk jdk = null;
        final Module module = getEffectiveModule();
        if (module != null) {
            jdk = ModuleRootManager.getInstance(module).getJdk();
        }
        if (jdk == null) {
            jdk = ProjectRootManager.getInstance(getProject()).getProjectJdk();
        }
        return jdk;
    }

    @Nullable
    Module getEffectiveModule() {
        assert myJdkChoice == JdkChoice.FROM_MODULE;

        Module module = getModule();
        if (module == null && myXsltFile != null) {
            final VirtualFile file = myXsltFile.getFile();
            if (file != null) {
                module = VfsUtil.getModuleForFile(getProject(), file);
            }
        }
        return module;
    }

    public boolean isGeneratedName() {
        return mySuggestedName != null;
    }

    public String suggestedName() {
        return mySuggestedName;
    }

    public static XsltRunConfiguration createFromFile(Project project, ConfigurationFactory factory, XmlFile file) {
        assert XsltSupport.isXsltFile(file);

        final VirtualFile virtualFile = file.getVirtualFile();
        assert virtualFile != null;

        final XsltRunConfiguration runConfiguration = new XsltRunConfiguration(project, factory, file.getName());

        runConfiguration.setXsltFile(virtualFile);

        final PsiFile[] associations = FileAssociationsManager.getInstance(file.getProject()).getAssociationsFor(file);
        if (associations.length > 0) {
            final VirtualFile assoc = associations[0].getVirtualFile();
            assert assoc != null;
            runConfiguration.setXmlInputFile(assoc);
        }

        final XmlDocument document = file.getDocument();
        assert document != null : "XSLT file without document?";
        
        final XmlTag rootTag = document.getRootTag();
        assert rootTag != null : "XSLT file without root element?";

        final XmlTag[] params = rootTag.findSubTags("param", XsltSupport.XSLT_NS);
        for (XmlTag param : params) {
            final String name = param.getAttributeValue("name");
            if (name != null) {
                final Pair<String, String> pair = Pair.create(name, null);
                runConfiguration.myParameters.add(pair);
            }
        }
        final XmlTag[] outputs = rootTag.findSubTags("output", XsltSupport.XSLT_NS);
        for (XmlTag output : outputs) {
            final String method = output.getAttributeValue("method");
            if ("xml".equals(method)) {
                runConfiguration.setFileType(StdFileTypes.XML);
            } else if ("html".equals(method)) {
                runConfiguration.setFileType(StdFileTypes.HTML);
            } else if ("text".equals(method)) {
                runConfiguration.setFileType(StdFileTypes.PLAIN_TEXT);
            }
        }

        return runConfiguration;
    }
}