/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.validation;

import org.intellij.lang.xpath.XPathFileType;
import org.intellij.lang.xpath.XPathLanguage;
import org.intellij.lang.xpath.psi.XPathVariableReference;
import org.intellij.lang.xpath.psi.impl.ResolveUtil;
import org.intellij.lang.xpath.xslt.XsltConfig;
import org.intellij.lang.xpath.xslt.XsltSupport;
import org.intellij.lang.xpath.xslt.psi.XsltApplyTemplates;
import org.intellij.lang.xpath.xslt.psi.XsltCallTemplate;
import org.intellij.lang.xpath.xslt.psi.XsltElementFactory;
import org.intellij.lang.xpath.xslt.psi.XsltNamedElement;
import org.intellij.lang.xpath.xslt.psi.XsltParameter;
import org.intellij.lang.xpath.xslt.psi.XsltTemplate;
import org.intellij.lang.xpath.xslt.psi.XsltTemplateInvocation;
import org.intellij.lang.xpath.xslt.psi.XsltWithParam;
import org.intellij.lang.xpath.xslt.quickfix.AddParameterFix;
import org.intellij.lang.xpath.xslt.quickfix.AddWithParamFix;
import org.intellij.lang.xpath.xslt.quickfix.RemoveParamFix;

import com.intellij.codeInsight.daemon.impl.HighlightInfo;
import com.intellij.codeInsight.daemon.impl.HighlightInfoType;
import com.intellij.codeInsight.daemon.impl.HighlightVisitor;
import com.intellij.codeInsight.daemon.impl.RefCountHolder;
import com.intellij.codeInsight.daemon.impl.analysis.HighlightInfoHolder;
import com.intellij.codeInsight.daemon.impl.quickfix.QuickFixAction;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiElementVisitor;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiNamedElement;
import com.intellij.psi.PsiRecursiveElementVisitor;
import com.intellij.psi.PsiReference;
import com.intellij.psi.PsiReferenceExpression;
import com.intellij.psi.search.LocalSearchScope;
import com.intellij.psi.search.SearchScope;
import com.intellij.psi.search.searches.ReferencesSearch;
import com.intellij.psi.xml.XmlAttribute;
import com.intellij.psi.xml.XmlAttributeValue;
import com.intellij.psi.xml.XmlTag;
import com.intellij.util.Processor;
import com.intellij.util.Query;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

public final class XsltDaemonHighlighter extends PsiElementVisitor implements HighlightVisitor, ProjectComponent,
        Cloneable, Processor<PsiReference> {

    private final boolean myEnabled;
    private final XPathLanguage myLanguage;
    private final XsltSupport myXsltSupport;
    private final XsltElementFactory myXsltElementFactory = XsltElementFactory.getInstance();
    private final Project myProject;

    private RefCountHolder myRefCountHolder;
    private HighlightInfoHolder myHighlightInfoHolder;


    XsltDaemonHighlighter(Project project) {
        myProject = project;
        myXsltSupport = XsltSupport.getInstance(project);
        myLanguage = (XPathLanguage)XPathFileType.XPATH.getLanguage();
        myEnabled = XsltConfig.getInstance().isEnabled();
    }

    public void init() {
        myRefCountHolder = null;
        myHighlightInfoHolder = null;
    }

    public HighlightVisitor clone() {
        try {
            return (HighlightVisitor)super.clone();
        } catch (CloneNotSupportedException e) {
            throw new Error(e);
        }
    }

    public void setRefCountHolder(RefCountHolder refCountHolder) {
        this.myRefCountHolder = refCountHolder;
    }

    public boolean suitableForFile(PsiFile psiFile) {
        return myEnabled && XsltSupport.isXsltFile(psiFile);
    }

    public void visitReferenceExpression(PsiReferenceExpression expression) {
    }

    public void visit(PsiElement element, HighlightInfoHolder highlightInfoHolder) {
        assert myHighlightInfoHolder == null;

        myHighlightInfoHolder = highlightInfoHolder;
        try {
            element.accept(this);
        } finally {
            myHighlightInfoHolder = null;
        }
    }

    public void visitXmlTag(final XmlTag tag) {
        final XmlAttribute nameAttr = tag.getAttribute("name", null);
        if (myRefCountHolder != null && nameAttr != null && XsltSupport.isVariableOrParam(tag)) {
            final XsltNamedElement instance = myXsltElementFactory.wrapElement(tag);
            checkDeclaration(instance, nameAttr.getValue(), true);

            if (!XsltSupport.isTopLevelElement(tag)) {
                final PsiNamedElement namedElement = (PsiNamedElement)instance;

                if (namedElement != null) {
                    final String name = namedElement.getName();
                    if (name != null && name.length() > 0) {
                        final boolean isAbstract;
                        if (namedElement instanceof XsltParameter) {
                            isAbstract = ((XsltParameter)namedElement).isAbstract();
                        } else {
                            isAbstract = false;
                        }
                        if (!isAbstract) {
                            myRefCountHolder.registerLocalDcl(namedElement);

                            // gotta do this, otherwise variables/params will become unused when editing inside the elements, e.g.
                            // when editing at <caret>, the variable would become marked as unused
                            //  <xsl:variable name="foo">
                            //     <xsl:value-of select="something<caret>"></xsl:value-of>
                            //  </xsl:variable>
                            //  <xsl:value-of select="$foo" />/

                            final SearchScope scope = new LocalSearchScope(tag.getParentTag());
                            final Query<PsiReference> refs = ReferencesSearch.search(instance, scope, false);
                            for (PsiReference reference : refs) {
                                reference.getElement().accept(this);
                            }
                        } else if (Arrays.asList(myRefCountHolder.getUnusedDcls()).contains(namedElement)) {
                            myRefCountHolder.registerLocallyReferenced(namedElement);
                        }
                    }
                }
            }
        } else if (nameAttr != null && XsltSupport.isTemplate(tag)) {
            final XsltTemplate tmpl = myXsltElementFactory.wrapElement(tag);
            checkDeclaration(tmpl, nameAttr.getValue(), false);
        } else if (XsltSupport.isTemplateCall(tag)) {
            final XsltCallTemplate call = myXsltElementFactory.wrapElement(tag);
            checkTemplateInvocation(call);
        } else if (XsltSupport.isApplyTemplates(tag)) {
            final XsltApplyTemplates call = myXsltElementFactory.wrapElement(tag);
            checkTemplateInvocation(call);
        }
    }

    private void checkTemplateInvocation(XsltTemplateInvocation call) {
        final XsltWithParam[] arguments = call.getArguments();

        final Map<String, XsltWithParam> argNames = new HashMap<String, XsltWithParam>();
        for (XsltWithParam arg : arguments) {
            final XmlAttribute attr = arg.getNameAttribute();
            if (attr != null) {
                final String name = attr.getValue();
                if (argNames.containsKey(name)) {
                    final PsiElement token = arg.getNameIdentifier();
                    assert token != null;
                    final HighlightInfo info = HighlightInfo.createHighlightInfo(HighlightInfoType.ERROR, token, "Duplicate Argument '" + name + "'");
                    myHighlightInfoHolder.add(info);
                }
                argNames.put(name, arg);
            }
        }

        if (call instanceof XsltCallTemplate) {
            final XsltCallTemplate ct = ((XsltCallTemplate)call);
            final PsiElement nameToken = ct.getNameIdentifier();
            final XsltTemplate template = ct.getTemplate();

            if (template != null) {
                if (nameToken != null) {
                    final XsltParameter[] parameters = template.getParameters();
                    for (XsltParameter parameter : parameters) {
                        if (!argNames.containsKey(parameter.getName()) && !parameter.hasDefault()) {
                            final HighlightInfo info = HighlightInfo.createHighlightInfo(HighlightInfoType.WARNING, nameToken, "Missing template parameter: " + parameter.getName());
                            myHighlightInfoHolder.add(info);
                            QuickFixAction.registerQuickFixAction(info, new AddWithParamFix(parameter, call.getTag()));
                        }
                    }
                }
                for (String s : argNames.keySet()) {
                    final XmlAttribute argAttribute = argNames.get(s).getNameAttribute();
                    assert argAttribute != null;

                    final XmlAttributeValue valueElement = argAttribute.getValueElement();
                    final PsiElement valueToken = XsltSupport.getAttValueToken(argAttribute);
                    if (valueToken != null && s.trim().length() > 0) {
                        if (template.getParameter(s) == null) {
                            final HighlightInfo info = HighlightInfo.createHighlightInfo(HighlightInfoType.WARNING, valueToken, "Superfluous template parameter: " + s);
                            myHighlightInfoHolder.add(info);
                            QuickFixAction.registerQuickFixAction(info, new AddParameterFix(s, template));
                            QuickFixAction.registerQuickFixAction(info, new RemoveParamFix(argNames.get(s).getTag(), s));
                        }
                    } else if (valueElement != null) {
                        myHighlightInfoHolder.add(HighlightInfo.createHighlightInfo(HighlightInfoType.ERROR, valueElement, "Parameter name expected"));
                    }
                }
            }
        }
    }

    private void checkDeclaration(final XsltNamedElement element, final String value, final boolean isVar) {
        final XmlTag tag = element.getTag();

        final PsiElement token = element.getNameIdentifier();
        if (value == null || value.length() == 0) {
            if (token != null) {
                myHighlightInfoHolder.add(HighlightInfo.createHighlightInfo(HighlightInfoType.ERROR, token, "Empty name not permitted"));
            } else {
                final XmlAttribute attribute = element.getNameAttribute();
                if (attribute != null) {
                    final XmlAttributeValue e = attribute.getValueElement();
                    if (e != null) {
                        myHighlightInfoHolder.add(HighlightInfo.createHighlightInfo(HighlightInfoType.ERROR, e, "Empty name not permitted"));
                    }
                }
            }
        } else if (!isLegalName(value)) {
            assert token != null;
            myHighlightInfoHolder.add(HighlightInfo.createHighlightInfo(HighlightInfoType.ERROR, token, "Illegal name"));
        } else {
            assert token != null;
            ResolveUtil.treeWalkUp(new DeclarationChecker(isVar, tag, value, myHighlightInfoHolder), tag);
        }
    }

    private boolean isLegalName(String value) {
        return myLanguage.getNamesValidator().isIdentifier(value, myProject);
    }

    public void visitXmlAttributeValue(XmlAttributeValue value) {
        if (myRefCountHolder != null) {
            final PsiElement parent = value.getParent();
            if (!(parent instanceof XmlAttribute)) {
                return;
            }
            final XmlAttribute attribute = (XmlAttribute)parent;

            if (XsltSupport.isXPathAttribute(attribute)) {
                final PsiFile[] files = myXsltSupport.getFiles(attribute);
                for (PsiFile file : files) {
                    scanLocalUsages(file);
                }
            }
        }
    }

    private void scanLocalUsages(PsiFile xpathFile) {
        xpathFile.accept(new PsiRecursiveElementVisitor() {
            public void visitElement(PsiElement element) {
                if (element instanceof XPathVariableReference) {
                    final PsiElement e = ((XPathVariableReference)element).resolve();
                    if (e instanceof PsiNamedElement) {
                        myRefCountHolder.registerLocallyReferenced((PsiNamedElement)e);
                    }
                }
                super.visitElement(element);
            }
        });
    }


    @NotNull
    public String getComponentName() {
        return "XsltHighlighter";
    }

    public void projectClosed() {
    }

    public void projectOpened() {
    }

    public void disposeComponent() {
    }

    public void initComponent() {
    }

    public boolean process(PsiReference ref) {
        ref.getElement().accept(this);
        return true;
    }
}