/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.validation.inspections;

import org.intellij.lang.xpath.validation.inspections.XPathInspection;
import org.intellij.lang.xpath.xslt.util.XsltCodeInsightUtil;

import com.intellij.codeInsight.intention.IntentionAction;
import com.intellij.codeInspection.LocalInspectionTool;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiWhiteSpace;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.*;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class InspectionUtil {
    static final Pattern SUPPRESSION_PATTERN = Pattern.compile("[ \t]*noinspection[ \t]+(\\w+(,[ \t]*\\w+)*)[ \t]*");

    private InspectionUtil() {
    }

    public static boolean isSuppressed(LocalInspectionTool tool, PsiElement element) {
        final XmlTag tag = PsiTreeUtil.getContextOfType(element, XmlTag.class, true);
        if (isSuppressedAt(tag, tool)) {
            return true;
        }
        final XmlTag tmpl = XsltCodeInsightUtil.getTemplateTag(element, true);
        if (isSuppressedAt(tmpl, tool)) {
            return true;
        }
        final XmlDocument document = PsiTreeUtil.getContextOfType(element, XmlDocument.class, true);
        if (document != null) {
            final XmlTag sheet = document.getRootTag();
            if (isSuppressedAt(sheet, tool)) {
                return true;
            }
        }
        return false;
    }

    private static boolean isSuppressedAt(PsiElement anchor, LocalInspectionTool tool) {
        if (anchor == null) return false;

        PsiElement prevSibling;
        if (!(anchor instanceof XmlComment)) {
            prevSibling = anchor.getPrevSibling();
            while (prevSibling instanceof PsiWhiteSpace || prevSibling instanceof XmlText) {
                prevSibling = prevSibling.getPrevSibling();
            }
        } else {
            prevSibling = anchor;
        }

        if (prevSibling instanceof XmlProlog) {
            if (prevSibling.getTextLength() > 0) {
                return isSuppressedAt(prevSibling.getLastChild(), tool);
            } else {
                return isSuppressedAt(prevSibling, tool);                
            }
        }
        if (prevSibling instanceof XmlComment) {
            final XmlComment comment = (XmlComment)prevSibling;
            final String text = XsltCodeInsightUtil.getCommentText(comment);
            if (text != null) {
                final Matcher matcher = SUPPRESSION_PATTERN.matcher(text);
                if (matcher.matches()) {
                    final String[] strings = matcher.group(1).split(",");
                    for (String s : strings) {
                        if (s.trim().equals(tool.getID())) return true;
                    }
                }
            }
        }
        return false;
    }

    public static List<? extends IntentionAction> getSuppressActions(ProblemDescriptor context, XPathInspection inspection) {
        final List<SuppressInspectionAction> actions = new ArrayList<SuppressInspectionAction>(3);

        final PsiElement element = context.getEndElement();
        final XmlTag tag = PsiTreeUtil.getContextOfType(element, XmlTag.class, true);
        actions.add(new SuppressInspectionAction(tag, inspection, "Suppress for Instruction"));
        final XmlTag tmpl = XsltCodeInsightUtil.getTemplateTag(element, true);
        actions.add(new SuppressInspectionAction(tmpl, inspection, "Suppress for Template"));
        final XmlDocument document = PsiTreeUtil.getContextOfType(element, XmlDocument.class, true);
        if (document != null) {
            final XmlTag sheet = document.getRootTag();
            actions.add(new SuppressInspectionAction(sheet, inspection, "Suppress for Stylesheet"));
        }
        return actions;
    }
}
