/*
 * Copyright 2005 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.lang.xpath.xslt.validation.inspections;

import org.intellij.lang.xpath.xslt.util.XsltCodeInsightUtil;

import com.intellij.codeInsight.intention.IntentionAction;
import com.intellij.psi.xml.XmlTag;
import com.intellij.psi.xml.XmlText;
import com.intellij.psi.xml.XmlProlog;
import com.intellij.psi.xml.XmlComment;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiWhiteSpace;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.codeInspection.LocalInspectionTool;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.editor.Editor;
import com.intellij.util.IncorrectOperationException;
import org.jetbrains.annotations.NotNull;

class SuppressInspectionAction implements IntentionAction {
    private final XmlTag myAnchor;
    private final LocalInspectionTool myTool;
    private final String myMsg;

    public SuppressInspectionAction(XmlTag anchor, LocalInspectionTool tool, String msg) {
        myAnchor = anchor;
        myTool = tool;
        myMsg = msg;
    }

    @NotNull
    public String getText() {
        return myMsg;
    }

    @NotNull
    public String getFamilyName() {
        return "Suppress Inspection";
    }

    public boolean isAvailable(Project project, Editor editor, PsiFile file) {
        return myAnchor != null;
    }

    public void invoke(Project project, Editor editor, PsiFile file) throws IncorrectOperationException {
        if (myAnchor == null) return;

        PsiElement prevSibling = myAnchor.getPrevSibling();
        while (prevSibling instanceof PsiWhiteSpace || prevSibling instanceof XmlText) {
            prevSibling = prevSibling.getPrevSibling();
        }
        if (prevSibling instanceof XmlProlog) {
            prevSibling = prevSibling.getLastChild();
            if (prevSibling != null && !(prevSibling instanceof XmlComment)) {
                prevSibling = PsiTreeUtil.getPrevSiblingOfType(prevSibling, XmlComment.class);
            }
        }
        if (prevSibling instanceof XmlComment) {
            final XmlComment comment = (XmlComment)prevSibling;
            final String text = XsltCodeInsightUtil.getCommentText(comment);
            if (text != null && InspectionUtil.SUPPRESSION_PATTERN.matcher(text).matches()) {
                final String s = text.trim() + ", " + myTool.getID();
                final XmlComment newComment = createComment(file, s);
                file.getManager().getCodeStyleManager().reformat(comment.replace(newComment));
            } else {
                addNoinspectionComment(file);
            }
        } else {
            addNoinspectionComment(file);
        }
    }

    private void addNoinspectionComment(PsiFile file) throws IncorrectOperationException {
        final XmlComment newComment = createComment(file, "noinspection " + myTool.getID());
        PsiElement parent = myAnchor.getParentTag();
        if (parent == null) {
            parent = PsiTreeUtil.getPrevSiblingOfType(myAnchor, XmlProlog.class);
            if (parent != null) {
                file.getManager().getCodeStyleManager().reformat(parent.add(newComment));
            }
        } else {
            file.getManager().getCodeStyleManager().reformat(parent.addBefore(newComment, myAnchor));
        }
    }

    @NotNull
    private static XmlComment createComment(PsiFile file, String s) throws IncorrectOperationException {
        final XmlTag element = file.getManager().getElementFactory().createTagFromText("<foo><!-- " + s + " --></foo>");
        final XmlComment newComment = PsiTreeUtil.getChildOfType(element, XmlComment.class);
        assert newComment != null;
        return newComment;
    }

    public boolean startInWriteAction() {
        return true;
    }
}
